# Generated by Django 4.1.3 on 2022-12-07 22:28

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("preferences", "0003_alter_basepreference_options_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="basepreference",
            options={
                "ordering": ["namespace", "key"],
                "verbose_name": "Settings",
                "verbose_name_plural": "Settings",
            },
        ),
        migrations.AlterModelOptions(
            name="scopedpreference",
            options={
                "ordering": ["base__namespace", "base__key", "scope"],
                "permissions": [
                    (
                        "change_own_user_scopedpreference",
                        "Can change own user's preferences",
                    ),
                    (
                        "change_own_tenant_scopedpreference",
                        "Can change own tenant's preferences",
                    ),
                    (
                        "change_own_group_scopedpreference",
                        "Can change own groups' preferences",
                    ),
                    (
                        "change_device_scopedpreference",
                        "Can change a device's preferences",
                    ),
                ],
                "verbose_name": "Scoped preferences",
                "verbose_name_plural": "Scoped preferences",
            },
        ),
    ]



from django.views.generic import DetailView, CreateView, ListView, TemplateView

from medux.core.models import Patient
from medux.core.views import MeduxBaseMixin


class PatientListView(MeduxBaseMixin, ListView):
    model = Patient
    permission_required = "can list patients"


class PatientFileView(MeduxBaseMixin, DetailView):
    # template_name = "patient_detail.html"
    permission_required = "can view patients"
    model = Patient

    # def get_context_data(self, *args, **kwargs):
    #     context = super().get_context_data(*args, **kwargs)
    #     return context


class NewPatientView(MeduxBaseMixin, CreateView):
    model = Patient
    permission_required = "can add patients"
    fields = ["names"]


class SettingsView(MeduxBaseMixin, TemplateView):
    template_name = "core/settings.html"
    permission_required = "core.canadmin"



from django.conf import settings
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.contrib.auth.views import LoginView as DjangoLoginView
from django.utils.translation import gettext_lazy as _
from django.views.generic import DetailView

from .multiple import MultipleFormsView, MultipleFormsMixin
from ..api.interfaces import IUserProfileSection, UseComponentMixin
from ..auth.forms import AuthenticationForm
from ..models import User, Tenant

__all__ = [MultipleFormsView, MultipleFormsMixin]


class LoginView(DjangoLoginView):
    """Login view with welcome message"""

    form_class = AuthenticationForm
    template_name = "registration/login.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["welcome_message"] = _("Medically welcome to {project_title}").format(
            project_title=settings.PROJECT_TITLE
        )
        return context

    def post(self, request, *args, **kwargs):
        return super().post(request, *args, **kwargs)


class FormActionsMixin:
    """Mixin to add to UpdateViews that enables additional action buttons.

    Each action is mapped to the corresponding method in the view, so a
    button named "delete" calls the ``delete()`` method.
    Actions available:
    * delete
    """

    # TODO: add wildcard buttons - is this a security risk?
    def post(self, request, *args, **kwargs):
        if request.POST.get("delete", ""):
            return self.delete(request, *args, **kwargs)
        else:
            return super().post(request, *args, **kwargs)


class DashboardMixin:
    """Add common needed dashboard attributes to the context.

    All views that live in the dashboard must inherit from this mixin.
    """

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update({"tenants": Tenant.objects.all().order_by("last_name")})
        return context


class UserUpdateView(
    PermissionRequiredMixin, DashboardMixin, UseComponentMixin, DetailView
):
    """The view for user data updates.

    This view doesn't have to be a UpdateView. It does not update data itself.
    The actual forms processing and user data updating is done in plugins.
    """

    model = User
    permission_required = "core.change_user"
    template_name = "common/user_change_form.html"
    components = [IUserProfileSection]


class UserProfileView(UserUpdateView):
    """A user update view, with the current authenticated user as model."""

    def has_permission(self):
        return self.request.user.is_authenticated

    def get_object(self, queryset=None):
        return self.request.user


class AutoFocusMixin:
    """Put the 'autofocus' attribute to a given field of a FormView.

    Set the autofocus attribute to a field of the form you are working
    with, and it will get the focus after loading the page.
    """

    # FIXME: move AutoFocusMixin to medux.common

    autofocus_field_name = None

    def get_form(self, form_class=None):
        """Return an instance of the form to be used in this view."""
        form = super().get_form(form_class)

        if self.autofocus_field_name in form.fields:
            form.fields[self.autofocus_field_name].widget.attrs.update(
                {"autofocus": True}
            )
        return form



from .common import *

# from .fhir import *
from .patient import *
from .datapacks import *
from .geo import *
from .medication import *
from .organizations import *
from .devices import *
from .observations import *



import graphene
from .models import User, Patient
from graphene_django import DjangoObjectType

from gdaps.graphene.api import IGrapheneSchema


# User
class UserType(DjangoObjectType):
    """MedUX's built-in User type"""

    class Meta:
        model = User


class UserQuery:
    users = graphene.List(UserType)

    @staticmethod
    def resolve_users(self, info, **kwargs):
        return User.objects.all()


class UserSchema(IGrapheneSchema):
    query = UserQuery


# Patients
class PatientType(DjangoObjectType):
    """A basic representation of a patient"""

    class Meta:
        model = Patient


class PatientQuery:
    patients = graphene.List(PatientType)

    @staticmethod
    def resolve_patients(self, info, **kwargs):
        return Patient.objects.all()


class PatientSchema(IGrapheneSchema):
    query = PatientQuery



#  MedUX - Open Source Electronical Medical Record
#  Copyright (c) 2022  Christian González
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.conf import settings
from django.contrib.auth.models import Group
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _

from medux.common.models import Tenant
from medux.preferences.definitions import Scope, KeyType, ScopeIcons
from medux.preferences.registry import PreferencesRegistry


def validate_lower(value):
    if not value == value.lower():
        raise ValidationError("%(value)s must be lowercase.", params={"value": value})


class BasePreference(models.Model):
    """Basic model which holds namespace and key of a setting.
    ScopedPreference use this model as FK, to prevent redundant storage of preferences."""

    class Meta:
        verbose_name = verbose_name_plural = _("Settings")
        ordering = ["namespace", "key"]

    namespace = models.CharField(max_length=25, validators=[validate_lower])
    key = models.CharField(max_length=255, validators=[validate_lower])

    def clean(self):
        """Make sure namespace and key are lowercase."""
        self.namespace = str(self.namespace).lower()
        self.key = str(self.key).lower()

    @classmethod
    def namespaces(cls) -> list[str]:
        """:return: a list of available namespaces."""
        # FIXME this is highly insufficient. distinct() would be better, but not available on SQLite/dev
        result = set()
        for s in cls.objects.order_by("namespace").values_list("namespace", flat=True):
            result.add(s)
        return list(result)

    @classmethod
    def keys(cls) -> list[tuple[str, str]]:
        """:return: a tuple[namespace,key] of all currently available keys."""

        return list(set((item.namespace, item.key) for item in cls.objects.all()))

    @property
    def icon(self):
        """:return: the preferences' icon name."""
        return PreferencesRegistry.icon(self.namespace, self.key)

    @property
    def key_type(self) -> KeyType:
        """:return: the setting's key type.

        This is a convenience method and gets it from the PreferencesRegistry."""
        return PreferencesRegistry.key_type(self.namespace, self.key)

    @property
    def help_text(self) -> str:
        """:return: the setting's help_text.
        This is a convenience method and gets it from the PreferencesRegistry."""
        return PreferencesRegistry.help_text(self.namespace, self.key)

    def __str__(self):
        return f"{self.namespace}.{self.key}"


class ScopedPreference(models.Model):
    """Model class for all scoped MedUX preferences.

    Preferences are generally saved as strings, but are interpreted at retrieval
    and cast into their correct types. ScopedPreference knows about
    ``str``, ``int``, ``bool``

    You can easily access preferences using CachedPreferences.get(namespace, key, scope, ...)
    """

    class Meta:
        verbose_name = verbose_name_plural = _("Scoped preferences")
        ordering = ["base__namespace", "base__key", "scope"]
        unique_together = [
            ["base", "scope", "tenant"],
            ["base", "scope", "group"],
            # TODO: ["base", "scope", "device"],
            ["base", "scope", "user"],
        ]
        permissions = [
            (
                "change_own_user_scopedpreference",
                _("Can change own user's preferences"),
            ),
            (
                "change_own_tenant_scopedpreference",
                _("Can change own tenant's preferences"),
            ),
            (
                "change_own_group_scopedpreference",
                _("Can change own groups' preferences"),
            ),
            ("change_device_scopedpreference", _("Can change a device's preferences")),
        ]

    base = models.ForeignKey(
        BasePreference, on_delete=models.CASCADE, related_name="scopedpreferences"
    )
    """The FK to the basic preferences fields like namespace, key."""

    tenant = models.ForeignKey(
        Tenant,
        verbose_name=_("Tenant"),
        on_delete=models.CASCADE,
        default=None,
        null=True,
        blank=True,
    )
    group = models.ForeignKey(
        Group, on_delete=models.CASCADE, default=None, null=True, blank=True
    )
    # device = models.ForeignKey(
    #     "Device", on_delete=models.CASCADE, default=None, null=True, blank=True
    # )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        default=None,
        null=True,
        blank=True,
    )

    scope = models.IntegerField(choices=Scope.choices)
    """The scope where this setting is valid."""

    value = models.CharField(max_length=255, null=True)
    """The value of the setting. NULL means the setting using this scope
    was deleted, and the next matching scope is used."""

    @property
    def namespace(self):
        """:return: The preferences' namespace"""
        return self.base.namespace

    @property
    def key(self):
        """:return: The preferences' key"""
        return self.base.key

    @property
    def icon(self):
        """:return: the preferences' icon name."""
        return self.base.icon

    @property
    def key_type(self) -> KeyType:
        """:return: the preferences's key type.

        This is a convenience method and gets it from the PreferencesRegistry."""
        return self.base.key_type

    @property
    def help_text(self) -> str:
        return self.base.help_text

    def clean(self):
        """Check if object FK is present if scope demands it."""

        # Make value a str
        if self.value is None:
            self.value = ""
        self.value = self.value.strip()
        # Scopes with their correct object FK
        if self.scope == Scope.USER:
            if not self.user:
                raise ValidationError(_("If scope is 'user', a user must be provided"))
        elif self.scope == Scope.DEVICE:
            if not self.device:
                raise ValidationError(
                    _("If scope is 'device', a device must be provided")
                )
        elif self.scope == Scope.GROUP:
            if not self.group:
                raise ValidationError(
                    _("If scope is 'group', a group must be provided")
                )
        elif self.scope == Scope.TENANT:
            if not self.tenant:
                raise ValidationError(
                    _("If scope is 'tenant', a tenant must be provided")
                )
        # TODO check if user/device/group is filled in if scope is one of them

    @classmethod
    def keys(cls) -> list[tuple[str, str]]:
        """:return: a Tuple[namespace,key] of all currently available keys."""

        return BasePreference.keys()

    @classmethod
    def namespaces(cls) -> list[str]:
        """:return: a list of available namespaces."""
        return BasePreference.namespaces()

    def __str__(self) -> str:
        # add user, tenant, group
        fk = ""
        if self.scope == Scope.USER:
            fk = f": '{self.user}'"
        elif self.scope == Scope.GROUP:
            fk = f": '{self.group}'"
        elif self.scope == Scope.TENANT:
            fk = f": '{self.tenant}'"
        a = f"{'.'.join([self.base.namespace, self.base.key])} [{Scope(self.scope).name}{fk}]: {self.value}"
        return a

    @classmethod
    def assure_exists(cls, namespace, key, scope) -> None:
        """Raises KeyError if given preferences are not registered."""
        if not PreferencesRegistry.exists(namespace, key, scope):
            raise KeyError(f"Setting {namespace}.{key}/{scope.name} is not registered.")

    def get_scope_icon(self) -> str:
        """:return: the icon name of the preferences' scope."""
        if not self.scope:
            return ""
        scope = Scope(self.scope)
        return ScopeIcons[scope.name].value if scope else ""

    def get_related_object(self):
        match self.scope:
            case Scope.TENANT:
                return self.tenant
            case Scope.USER:
                return self.user
            case Scope.DEVICE:
                return self.device
            case Scope.GROUP:
                return self.group
            case Scope.VENDOR:
                return None
                # TODO: maybe return Vendor name here?



from enum import Enum

from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _

from ..fields import CodeField
from medux.core.models import PackageDataModel
from medux.common.models import BaseModel


class DeviceType(PackageDataModel):
    pass


class Device(BaseModel):
    class Status(Enum):
        ACTIVE = "active"
        INACTIVE = "inactive"
        ENTERED_IN_ERROR = "entered_in_error"
        UNKNOWN = "unknown"

    class StatusReason(Enum):
        ONLINE = "online"
        PAUSED = "paused"
        STANDBY = "standby"
        OFFLINE = "offline"
        NOT_READY = "not_ready"
        TRANSDUC_DISCONNECTED = "transduc_discon"
        HARDWARE_DISCONNECTED = "hardware__discon"
        OFF = "off"

    status = CodeField(
        terminology_binding=Status,
        default="active",
        help_text=_("Status of the device availability"),
    )
    statusReason = CodeField(
        terminology_binding=StatusReason,
        default="online",
        help_text=_("reason for the status of the device availability"),
    )
    distinct_identifier = models.CharField(
        max_length=255, blank=True, help_text=_("The distinct identification string")
    )
    manufacturer = models.CharField(
        max_length=255, blank=True, help_text=_("Name of device manufacturer")
    )
    manufacture_date = models.DateField(
        blank=True, null=True, help_text=_("Date when the device was made")
    )
    expiration_date = models.DateField(
        blank=True,
        null=True,
        help_text=_("Date and time of expiry of this device (if applicable)"),
    )
    lot_number = models.CharField(
        max_length=255,
        blank=True,
        help_text=_("Lot number assigned by the manufacturer"),
    )
    serial_number = models.CharField(
        max_length=255,
        blank=True,
        help_text=_("Serial number assigned by the manufacturer"),
    )
    device_name = models.CharField(
        max_length=255,
        blank=True,
        help_text=_("The name of the device as given by the manufacturer"),
    )
    model_number = models.CharField(
        max_length=255,
        blank=True,
        help_text=_("The model number for the device"),
    )
    part_number = models.CharField(
        max_length=255,
        blank=True,
        help_text=_("The part number for the device"),
    )

    type = models.ForeignKey(
        DeviceType,
        blank=True,
        null=True,
        help_text=_("The kind or type of device"),
        on_delete=models.PROTECT,
    )

    # specialization
    # version
    # property

    patient = models.ForeignKey(
        "Patient",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        help_text=_("Patient to whom Device is affixed"),
    )

    owner = models.ForeignKey(
        "Organization",
        null=True,
        on_delete=models.SET_NULL,
        help_text=_("Organization responsible for device"),
    )
    contact = models.ManyToManyField("ContactPoint", blank=True)
    location = models.CharField(max_length=255, blank=True)
    url = models.URLField(blank=True)
    parent = models.ForeignKey(
        "Device", blank=True, null=True, on_delete=models.CASCADE
    )

    def __str__(self):
        return f"{self.device_name} (SN: {self.serial_number or _('n/a')})"

    def clean(self):
        if not self.device_name and not self.serial_number:
            raise ValidationError(
                _("A device must have either a device name or a serial number.")
            )



from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponseRedirect
from django.views.generic.base import ContextMixin, TemplateResponseMixin
from django.views.generic.edit import ProcessFormView

# The following code is licensed under the MIT licence
# https://github.com/Valian/django-multiple-forms


class MultipleFormsMixin(ContextMixin):
    """
    A mixin that provides a way to show and handle multiple forms in a request.
    It's almost fully-compatible with regular FormsMixin
    """

    initial = {}
    forms_classes = []
    success_url = None
    prefix = None
    active_form_keyword = "selected_form"

    def get_initial(self):
        """
        Returns the initial data to use for forms on this view.
        """
        return self.initial.copy()

    def get_prefix(self):
        """
        Returns the prefix to use for forms on this view
        """
        return self.prefix

    def get_forms_classes(self):
        """
        Returns the forms classes to use in this view. Can be dynamic.
        """
        return self.forms_classes

    def get_active_form_number(self):
        """
        Returns submitted form index in available forms list
        """
        if self.request.method in ("POST", "PUT"):
            try:
                return int(self.request.POST[self.active_form_keyword])
            except (KeyError, ValueError):
                raise ImproperlyConfigured(
                    "You must include hidden field with field index in every form! Example:"
                )

    def get_forms(self, active_form=None):
        """
        Returns instances of the forms to be used in this view.
        Includes provided `active_form` in forms list.
        """
        all_forms_classes = self.get_forms_classes()
        all_forms = [
            form_class(**self.get_form_kwargs()) for form_class in all_forms_classes
        ]
        if active_form:
            active_form_number = self.get_active_form_number()
            all_forms[active_form_number] = active_form
        return all_forms

    def get_form(self):
        """
        Returns active form. Works only on `POST` and `PUT`, otherwise returns None.
        """
        active_form_number = self.get_active_form_number()
        if active_form_number is not None:
            all_forms_classes = self.get_forms_classes()
            active_form_class = all_forms_classes[active_form_number]
            return active_form_class(**self.get_form_kwargs(is_active=True))

    def get_form_kwargs(self, is_active=False):
        """
        Returns the keyword arguments for instantiating the form.
        """
        kwargs = {
            "initial": self.get_initial(),
            "prefix": self.get_prefix(),
        }

        if is_active:
            kwargs.update(
                {
                    "data": self.request.POST,
                    "files": self.request.FILES,
                }
            )
        return kwargs

    def get_success_url(self):
        """
        Returns the supplied success URL.
        """
        if self.success_url:
            # Forcing possible reverse_lazy evaluation
            url = self.success_url
        else:
            raise ImproperlyConfigured("No URL to redirect to. Provide a success_url.")
        return url

    def form_valid(self, form):
        """
        If the form is valid, redirect to the supplied URL.
        """
        return HttpResponseRedirect(self.get_success_url())

    def form_invalid(self, form):
        """
        If the form is invalid, re-render the context data with the
        data-filled forms and errors.
        """
        return self.render_to_response(self.get_context_data(active_form=form))

    def get_context_data(self, **kwargs):
        """
        Insert the forms into the context dict.
        """
        if "forms" not in kwargs:
            kwargs["forms"] = self.get_forms(kwargs.get("active_form"))
        return super(MultipleFormsMixin, self).get_context_data(**kwargs)


class MultipleFormsView(TemplateResponseMixin, MultipleFormsMixin, ProcessFormView):
    pass



import uuid
from io import BytesIO

import phonenumbers
from django.contrib.auth.models import AbstractUser
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models.enums import TextChoices
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from PIL import Image, ImageDraw

from medux.common.tools import resize_image


class AutoDateTimeField(models.DateTimeField):
    def pre_save(self, model_instance, add):
        return timezone.now()


class SingletonModel(models.Model):
    """A base model that only can haz one instance."""

    class Meta:
        abstract = True

    @classmethod
    def get_instance(cls):
        """:return: the current instance if available, else an empty instance."""
        if cls.objects.exists():
            return cls.objects.first()
        else:
            return cls()

    def save(self, *args, **kwargs):
        if not self.pk and Vendor.objects.exists():
            raise ValidationError(
                f"There can be only one {self.__class__.__name__} instance."
            )
        super().save(*args, **kwargs)


class CreatedModifiedModel(models.Model):
    """A simple mixin for model classes that need to have created/modified fields."""

    created = models.DateTimeField(editable=False, default=timezone.now)
    modified = AutoDateTimeField(default=timezone.now)

    class Meta:
        abstract = True


class SoftDeletionQuerySet(models.QuerySet):
    def delete(self):
        return super().update(deleted_at=timezone.now())

    def hard_delete(self):
        return super().delete()

    def alive(self):
        return self.filter(deleted_at=None)

    def dead(self):
        return self.exclude(deleted_at=None)


class SoftDeletionManager(models.Manager):
    """Custom Django Manager for soft-delete queries.

    This object manager transparently only fetches objects from the
    database that are not soft-deleted."""

    # https://medium.com/@adriennedomingus/soft-deletion-in-django-e4882581c340
    def __init__(self, *args, **kwargs):
        self.alive_only = kwargs.pop("alive_only", True)
        super().__init__(*args, **kwargs)

    def get_queryset(self):
        if self.alive_only:
            return SoftDeletionQuerySet(self.model).filter(deleted_at=None)
        return SoftDeletionQuerySet(self.model)

    def hard_delete(self):
        return self.get_queryset().hard_delete()


class BaseModel(CreatedModifiedModel):
    """An abstract base class with common used functions.

    Every relevant model that needs soft_deletion in MedUX should inherit BaseModel.

    It provides:
    * basic created/modified timestamps for auditing
    * a soft delete functionality: Deleted items are just marked as deleted.
    """

    deleted_at = models.DateTimeField(
        editable=False, blank=True, null=True, default=None
    )

    row_version = models.PositiveIntegerField(editable=False, default=0)

    # The standard manager only returns not-soft-deleted objects
    objects = SoftDeletionManager()

    # The all_objects Manager returns ALL objects, even soft-deleted ones
    all_objects = SoftDeletionManager(alive_only=False)

    class Meta:
        abstract = True

    def delete(self, using=None, keep_parents=False):
        self.deleted_at = timezone.now()
        self.save()

    def hard_delete(self):
        super().delete()


class SexChoices(TextChoices):
    MALE = "m", _("male")
    FEMALE = "f", _("female")


class Tenant(models.Model):
    """A MedUX tenant, like an MD who "owns" a homepage, or a practice
    owner with a MedUX appliance."""

    class Meta:
        verbose_name = _("Tenant")
        verbose_name_plural = _("Tenants")

    uuid = models.UUIDField(default=uuid.uuid4, unique=True)

    title = models.CharField(_("Title"), max_length=50, blank=True)
    first_name = models.CharField(_("First name"), max_length=255)
    last_name = models.CharField(_("Last name"), max_length=255)
    sex = models.CharField(_("Sex"), max_length=1, choices=SexChoices.choices)
    address = models.CharField(_("Address"), max_length=255)
    phone = models.CharField(_("Phone"), max_length=30, blank=True)
    email = models.EmailField(
        _("Email"), unique=True, default=None, blank=True, null=True
    )
    #: The picture/logo of the tenant. If None, the initials are used to create a picture
    picture = models.ImageField(_("Picture"), blank=True, null=True)

    def get_picture_from_initials(self):
        # TODO: make colers configurable
        canvas = Image.new("RGB", (128, 128), "grey")

        draw = ImageDraw.Draw(canvas)
        # font = ImageFont.truetype("FreeMono.ttf", 48)
        draw.text(
            (4, 4),
            self.initials(),
            # font=font,
            fill=(200, 50, 50),
        )
        blob = BytesIO()
        canvas.save(blob, "JPEG")  # TODO: png
        del blob
        return canvas

    @property
    def name(self):
        # TODO: if user exists, take his name.
        title = f"{self.title} " if self.title else ""
        return f"{title}{self.last_name}, {self.first_name}"

    def __str__(self):
        return f"{self.name}"

    def natural_key(self):
        return self.first_name, self.last_name

    def get_absolute_url(self):
        return reverse("tenant:detail", kwargs={"pk": self.pk})
        # FIXME: dashboard is not available here...

    def initials(self):
        return f"{self.last_name[0].upper()}{self.first_name[0].upper()}"


class User(AbstractUser):
    """Custom MedUX User model.

    A user usually belongs to a tenant, except for e.g. admin users.
    """

    title = models.CharField(max_length=25, blank=True, null=True)
    # FIXME: Admin user should belong to Admin tenant - no null necessary here
    tenant = models.ForeignKey(Tenant, on_delete=models.CASCADE, null=True)

    def __str__(self) -> str:
        title = f"{self.title} " if self.title else ""
        if self.first_name and self.last_name:
            return f"{title}{self.first_name} {self.last_name}"
        else:
            return self.username

    def natural_key(self) -> str:
        return self.username

    def get_absolute_url(self) -> str:
        """Get url for user's detail view.
        :returns str: URL for user detail."""
        return reverse("user:detail", kwargs={"pk": self.pk})

    avatar = models.ImageField(upload_to="user_avatars/", null=True)

    @property
    def job_title(self):
        return "Job title"  # TODO: add job title

    def save(self, commit=True, *args, **kwargs):
        """Resizes avatar image before saving."""
        if commit:
            if self.avatar:
                resize_image(self.avatar, 250, 250)
            super().save(*args, **kwargs)


class TenantModelMixin(models.Model):
    """A mixin that can be added to a model to mark it as belonging to a tenant.

    It adds a ``tenant`` ForeignKey to the model."""

    class Meta:
        abstract = True

    tenant = models.ForeignKey(
        Tenant,
        verbose_name=_("Tenant"),
        on_delete=models.CASCADE,
    )


# class OptionalTenantModelMixin(models.Model):
#     """A mixin that can be added to a model to mark it as
#     optionally belonging to a tenant.
#
#     It adds a (nullable) ``tenant`` ForeignKey to the model."""
#
#     class Meta:
#         abstract = True
#
#     tenant = models.ForeignKey(
#         Tenant,
#         verbose_name=_("Tenant"),
#         on_delete=models.CASCADE,
#         blank=True,
#         null=True,
#     )


# class TenantGroup(TenantModelMixin):
#     """A group within a tenant that has certain rights."""


def validate_phonenumber(number: str) -> bool:
    """return True iv given number is a possible phone number"""
    z = phonenumbers.parse(number)
    return phonenumbers.is_possible_number(z)


class Vendor(SingletonModel):
    """The vendor that is responsible for this MedUX appliance."""

    # TODO find an easy way to deploy data in Vendor model
    # maybe in a preferences.toml file?

    name = models.CharField(max_length=255)
    address = models.CharField(max_length=255)
    zip = models.CharField(max_length=10)
    city = models.CharField(max_length=50)
    phone = models.CharField(max_length=80, validators=[validate_phonenumber])
    email = models.EmailField(blank=True)

    def __str__(self):
        return self.name



from django.utils.translation import gettext_lazy as _

from gdaps.api import PluginConfig

from . import __version__
from medux.common.api import MeduxPluginAppConfig


class CoreConfig(MeduxPluginAppConfig):
    """MedUX Core Plugin"""

    name = "medux.core"
    groups_permissions = {
        "Users": {"core.Patient": ["view"]},
        "Patient managers": {"core.Patient": ["add", "change", "delete"]},
    }

    class PluginMeta:
        verbose_name = _("MedUX Core")
        author = "Christian González"
        author_email = "christian.gonzalez@nerdocs.at"
        vendor = "nerdocs"
        description = _("Medux Core Plugin")
        category = _("Core")
        visible = True
        version = __version__

    def ready(self):
        # This function is called after the app and all models are loaded.
        #
        # You can do some initialization here, but beware: it should rather
        # return fast, as it is called at each Django start, even on
        # management commands (makemigrations/migrate etc.).
        #
        # Avoid interacting with the database especially 'save' operations,
        # if you don't *really* have to."""

        try:
            from . import signals
        except ImportError:
            pass



from django.db import models
from django.utils.translation import gettext_lazy as _


class Scope(models.IntegerChoices):
    VENDOR = 1, _("Vendor")
    TENANT = 2, _("Tenant")
    GROUP = 3, _("Group")
    DEVICE = 4, _("Device")
    USER = 5, _("User")


class ScopeIcons(models.TextChoices):
    VENDOR = "box2"  # -fill
    TENANT = "person-badge"  # -fill
    GROUP = "people"  # -fill
    DEVICE = "display"  # -fill
    USER = "person"  # -fill


class KeyType(models.TextChoices):
    BOOLEAN = "bool"
    STRING = "str"  # short string
    INTEGER = "int"
    TEXT = "text"  # long text



# Generated by Django 4.0.6 on 2022-07-28 14:49

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import medux.common.models
import medux.core.fields
import medux.core.models.devices
import medux.core.models.geo
import medux.core.models.medication
import medux.core.models.observations
import uuid


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("core", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Address",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "modified",
                    medux.common.models.AutoDateTimeField(
                        default=django.utils.timezone.now
                    ),
                ),
                (
                    "deleted_at",
                    models.DateTimeField(
                        blank=True, default=None, editable=False, null=True
                    ),
                ),
                ("row_version", models.PositiveIntegerField(default=0, editable=False)),
                ("street", models.CharField(blank=True, max_length=255)),
                (
                    "aux_street",
                    models.CharField(
                        blank=True,
                        help_text="hints like postal box directions in CA, or c/o, etc.",
                        max_length=255,
                    ),
                ),
                (
                    "number",
                    models.CharField(
                        blank=True,
                        help_text="Street number, evtl. with additions like stair number, etc.",
                        max_length=50,
                    ),
                ),
                ("subunit", models.CharField(blank=True, max_length=255)),
                ("postcode", models.CharField(blank=True, max_length=10)),
                ("city", models.CharField(blank=True, max_length=255)),
                (
                    "addendum",
                    models.CharField(
                        blank=True,
                        help_text="Any additional information that did not fit anywhere else",
                        max_length=255,
                    ),
                ),
                (
                    "location",
                    models.CharField(
                        blank=True,
                        help_text="The exact location of this address in latitude-longitude",
                        max_length=255,
                    ),
                ),
                ("period_start", models.DateTimeField(blank=True, null=True)),
                ("period_end", models.DateTimeField(blank=True, null=True)),
                ("weight", models.PositiveIntegerField(default=1)),
            ],
            options={
                "verbose_name_plural": "Addresses",
                "ordering": ["weight"],
            },
        ),
        migrations.CreateModel(
            name="AddressMapper",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "address_id",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT, to="core.address"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="AddressType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name="AdministrativeGender",
            fields=[
                (
                    "tag",
                    models.CharField(max_length=2, primary_key=True, serialize=False),
                ),
                ("label", models.CharField(max_length=50)),
                ("sort_weight", models.IntegerField(default=0)),
                ("comment", models.CharField(blank=True, max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name="CarePlan",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "modified",
                    medux.common.models.AutoDateTimeField(
                        default=django.utils.timezone.now
                    ),
                ),
                (
                    "deleted_at",
                    models.DateTimeField(
                        blank=True, default=None, editable=False, null=True
                    ),
                ),
                ("row_version", models.PositiveIntegerField(default=0, editable=False)),
                (
                    "status",
                    medux.core.fields.CodeField(
                        max_length=32,
                        terminology_binding=medux.core.models.medication.CarePlan.Status,
                    ),
                ),
                (
                    "intent",
                    medux.core.fields.CodeField(
                        max_length=32,
                        terminology_binding=medux.core.models.medication.CarePlan.Intents,
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        blank=True,
                        help_text="Human-friendly name for the care plan",
                        max_length=255,
                    ),
                ),
                (
                    "description",
                    models.CharField(
                        blank=True, help_text="Summary of plan", max_length=255
                    ),
                ),
                ("period_start", models.DateTimeField(blank=True, null=True)),
                ("period_end", models.DateTimeField(blank=True, null=True)),
                (
                    "note",
                    models.TextField(blank=True, help_text="Comments about the plan"),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="City",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                (
                    "maintainer",
                    models.SmallIntegerField(
                        choices=[(1, "User"), (10, "Admin"), (99, "Vendor")], default=1
                    ),
                ),
                ("name", models.CharField(max_length=100)),
            ],
            options={
                "verbose_name_plural": "Cities",
            },
        ),
        migrations.CreateModel(
            name="ContactPoint",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "modified",
                    medux.common.models.AutoDateTimeField(
                        default=django.utils.timezone.now
                    ),
                ),
                (
                    "deleted_at",
                    models.DateTimeField(
                        blank=True, default=None, editable=False, null=True
                    ),
                ),
                ("row_version", models.PositiveIntegerField(default=0, editable=False)),
                (
                    "system",
                    models.CharField(
                        choices=[
                            ("phone", "Telefon"),
                            ("fax", "Fax"),
                            ("email", "E-Mail-Adresse"),
                            ("pager", "Pager"),
                            ("url", "URL"),
                            ("sms", "SMS"),
                            ("other", "Andere"),
                        ],
                        max_length=20,
                    ),
                ),
                ("value", models.CharField(blank=True, max_length=255)),
                (
                    "use",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("home", "Privat"),
                            ("work", "Geschäftlich"),
                            ("temp", "Termporär"),
                            ("old", "Veraltet"),
                            ("mobile", "Mobil"),
                        ],
                        max_length=20,
                    ),
                ),
                ("weight", models.PositiveIntegerField(blank=True, default=1)),
                ("period_start", models.DateTimeField(blank=True, null=True)),
                ("period_end", models.DateTimeField(blank=True, null=True)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Country",
            fields=[
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                (
                    "maintainer",
                    models.SmallIntegerField(
                        choices=[(1, "User"), (10, "Admin"), (99, "Vendor")], default=1
                    ),
                ),
                (
                    "code",
                    models.CharField(
                        help_text="ISO 3166-1 code",
                        max_length=2,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Official English short name", max_length=255
                    ),
                ),
                ("flag", models.CharField(blank=True, max_length=5)),
            ],
            options={
                "verbose_name_plural": "Countries",
            },
            bases=(medux.core.models.geo.UpperCodeMixin, models.Model),
        ),
        migrations.CreateModel(
            name="CourseOfTherapy",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                (
                    "maintainer",
                    models.SmallIntegerField(
                        choices=[(1, "User"), (10, "Admin"), (99, "Vendor")], default=1
                    ),
                ),
                ("code", models.CharField(max_length=50)),
                ("display", models.CharField(max_length=50)),
                ("definition", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="DataPack",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "modified",
                    medux.common.models.AutoDateTimeField(
                        default=django.utils.timezone.now
                    ),
                ),
                (
                    "deleted_at",
                    models.DateTimeField(
                        blank=True, default=None, editable=False, null=True
                    ),
                ),
                ("row_version", models.PositiveIntegerField(default=0, editable=False)),
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                (
                    "name",
                    models.CharField(
                        help_text="Name for this data pack (computer friendly)",
                        max_length=100,
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        help_text="Name for this data pack (human friendly)",
                        max_length=255,
                    ),
                ),
                (
                    "description",
                    medux.core.fields.MarkdownField(
                        help_text="Natural description, in MarkDown", max_length=255
                    ),
                ),
                ("license", models.CharField(max_length=20)),
                ("version", models.CharField(max_length=25)),
                ("experimental", models.BooleanField(default=False)),
                ("publisher", models.CharField(blank=True, max_length=255)),
                ("model", models.CharField(max_length=255)),
                ("downloaded", models.BooleanField(default=False)),
                ("installed", models.BooleanField(default=False)),
                ("data_file", models.FileField(upload_to="datapacks")),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Department",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                (
                    "maintainer",
                    models.SmallIntegerField(
                        choices=[(1, "User"), (10, "Admin"), (99, "Vendor")], default=1
                    ),
                ),
                ("name", models.CharField(max_length=255)),
                ("telecom", models.ManyToManyField(to="core.contactpoint")),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Device",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "modified",
                    medux.common.models.AutoDateTimeField(
                        default=django.utils.timezone.now
                    ),
                ),
                (
                    "deleted_at",
                    models.DateTimeField(
                        blank=True, default=None, editable=False, null=True
                    ),
                ),
                ("row_version", models.PositiveIntegerField(default=0, editable=False)),
                (
                    "status",
                    medux.core.fields.CodeField(
                        default="active",
                        help_text="Status of the device availability",
                        max_length=32,
                        terminology_binding=medux.core.models.devices.Device.Status,
                    ),
                ),
                (
                    "statusReason",
                    medux.core.fields.CodeField(
                        default="online",
                        help_text="reason for the status of the device availability",
                        max_length=32,
                        terminology_binding=medux.core.models.devices.Device.StatusReason,
                    ),
                ),
                (
                    "distinct_identifier",
                    models.CharField(
                        blank=True,
                        help_text="The distinct identification string",
                        max_length=255,
                    ),
                ),
                (
                    "manufacturer",
                    models.CharField(
                        blank=True,
                        help_text="Name of device manufacturer",
                        max_length=255,
                    ),
                ),
                (
                    "manufacture_date",
                    models.DateField(
                        blank=True, help_text="Date when the device was made", null=True
                    ),
                ),
                (
                    "expiration_date",
                    models.DateField(
                        blank=True,
                        help_text="Date and time of expiry of this device (if applicable)",
                        null=True,
                    ),
                ),
                (
                    "lot_number",
                    models.CharField(
                        blank=True,
                        help_text="Lot number assigned by the manufacturer",
                        max_length=255,
                    ),
                ),
                (
                    "serial_number",
                    models.CharField(
                        blank=True,
                        help_text="Serial number assigned by the manufacturer",
                        max_length=255,
                    ),
                ),
                (
                    "device_name",
                    models.CharField(
                        blank=True,
                        help_text="The name of the device as given by the manufacturer",
                        max_length=255,
                    ),
                ),
                (
                    "model_number",
                    models.CharField(
                        blank=True,
                        help_text="The model number for the device",
                        max_length=255,
                    ),
                ),
                (
                    "part_number",
                    models.CharField(
                        blank=True,
                        help_text="The part number for the device",
                        max_length=255,
                    ),
                ),
                ("location", models.CharField(blank=True, max_length=255)),
                ("url", models.URLField(blank=True)),
                ("contact", models.ManyToManyField(blank=True, to="core.contactpoint")),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="DeviceType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                (
                    "maintainer",
                    models.SmallIntegerField(
                        choices=[(1, "User"), (10, "Admin"), (99, "Vendor")], default=1
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Dosage",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "dose_rate_numerator_value",
                    models.DecimalField(
                        blank=True, decimal_places=1, max_digits=1, null=True
                    ),
                ),
                (
                    "dose_rate_numerator_unit",
                    medux.core.fields.UnitField(blank=True, max_length=10),
                ),
                (
                    "dose_rate_numerator_system",
                    models.CharField(blank=True, max_length=255),
                ),
                (
                    "dose_rate_numerator_comparator",
                    models.CharField(
                        blank=True,
                        choices=[("<", "<"), ("<=", "<="), (">=", ">="), (">", ">")],
                        max_length=2,
                    ),
                ),
                (
                    "dose_rate_denominator_value",
                    models.DecimalField(decimal_places=1, max_digits=1),
                ),
                (
                    "dose_rate_denominator_unit",
                    medux.core.fields.UnitField(blank=True, max_length=10),
                ),
                (
                    "dose_rate_denominator_system",
                    models.CharField(blank=True, max_length=255),
                ),
                (
                    "dose_rate_denominator_comparator",
                    models.CharField(
                        blank=True,
                        choices=[("<", "<"), ("<=", "<="), (">=", ">="), (">", ">")],
                        max_length=2,
                    ),
                ),
                (
                    "text",
                    models.CharField(
                        help_text="Free text dosage instructions e.g. '1x/day'",
                        max_length=255,
                    ),
                ),
                (
                    "additional_instructions",
                    models.CharField(
                        blank=True,
                        help_text="Supplemental instruction or warnings to the patient - e.g. 'with meals', 'may cause drowsiness'",
                        max_length=255,
                    ),
                ),
                (
                    "patient_instructions",
                    models.CharField(
                        blank=True,
                        help_text="Instructions in terms that are understood by the patient",
                        max_length=255,
                    ),
                ),
                (
                    "as_needed",
                    models.BooleanField(
                        default=False, help_text="Take 'as needed' (for ...)"
                    ),
                ),
                (
                    "dose_type",
                    medux.core.fields.CodeField(
                        default="ordered",
                        max_length=32,
                        terminology_binding=medux.core.models.medication.Dosage.DoseAndRateType,
                    ),
                ),
                (
                    "dose_quantity_value",
                    models.DecimalField(
                        blank=True, decimal_places=2, max_digits=10, null=True
                    ),
                ),
                (
                    "dose_quantity_unit",
                    medux.core.fields.UnitField(blank=True, max_length=10),
                ),
                ("dose_quantity_system", models.CharField(blank=True, max_length=255)),
                (
                    "dose_quantity_comparator",
                    models.CharField(
                        blank=True,
                        choices=[("<", "<"), ("<=", "<="), (">=", ">="), (">", ">")],
                        max_length=2,
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Encounter",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "modified",
                    medux.common.models.AutoDateTimeField(
                        default=django.utils.timezone.now
                    ),
                ),
                (
                    "deleted_at",
                    models.DateTimeField(
                        blank=True, default=None, editable=False, null=True
                    ),
                ),
                ("row_version", models.PositiveIntegerField(default=0, editable=False)),
                ("date", models.DateTimeField(default=django.utils.timezone.now)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="EpisodeOfCare",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "modified",
                    medux.common.models.AutoDateTimeField(
                        default=django.utils.timezone.now
                    ),
                ),
                (
                    "deleted_at",
                    models.DateTimeField(
                        blank=True, default=None, editable=False, null=True
                    ),
                ),
                ("row_version", models.PositiveIntegerField(default=0, editable=False)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="EpisodeStatus",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                (
                    "maintainer",
                    models.SmallIntegerField(
                        choices=[(1, "User"), (10, "Admin"), (99, "Vendor")], default=1
                    ),
                ),
                ("code", models.CharField(max_length=50)),
                ("display", models.CharField(max_length=50)),
                ("definition", models.CharField(max_length=255)),
            ],
            options={
                "verbose_name_plural": "Episode Statuses",
            },
        ),
        migrations.CreateModel(
            name="EpisodeStatusHistory",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("period_start", models.DateTimeField(blank=True, null=True)),
                ("period_end", models.DateTimeField(blank=True, null=True)),
                (
                    "episode",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="core.episodeofcare",
                    ),
                ),
                (
                    "status",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="core.episodestatus",
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Episode Status Histories",
            },
        ),
        migrations.CreateModel(
            name="Firstname",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                (
                    "maintainer",
                    models.SmallIntegerField(
                        choices=[(1, "User"), (10, "Admin"), (99, "Vendor")], default=1
                    ),
                ),
                ("name", models.CharField(max_length=50)),
                (
                    "sex",
                    models.CharField(
                        choices=[
                            ("M", "Male"),
                            (
                                "1M",
                                "male if first part of name, otherwise mostly female",
                            ),
                            ("?M", "mostly male"),
                            ("F", "Female"),
                            (
                                "1F",
                                "female if first part of name, otherwise mostly male",
                            ),
                            ("?F", "mostly female"),
                            ("?", "unisex"),
                        ],
                        max_length=2,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Ingredient",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Language",
            fields=[
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                (
                    "maintainer",
                    models.SmallIntegerField(
                        choices=[(1, "User"), (10, "Admin"), (99, "Vendor")], default=1
                    ),
                ),
                (
                    "code",
                    models.CharField(
                        help_text="ISO 639-1 language code",
                        max_length=2,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="ISO 639-1 English language name", max_length=50
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=(medux.core.models.geo.UpperCodeMixin, models.Model),
        ),
        migrations.CreateModel(
            name="Medication",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                (
                    "maintainer",
                    models.SmallIntegerField(
                        choices=[(1, "User"), (10, "Admin"), (99, "Vendor")], default=1
                    ),
                ),
                (
                    "code",
                    models.CharField(
                        help_text="The medication code within this coding system",
                        max_length=64,
                    ),
                ),
                (
                    "status",
                    medux.core.fields.CodeField(
                        default="active",
                        max_length=32,
                        terminology_binding=medux.core.models.medication.Medication.Status,
                    ),
                ),
                (
                    "amount_value",
                    models.DecimalField(
                        decimal_places=3,
                        help_text="Amount of drug in package",
                        max_digits=10,
                    ),
                ),
                (
                    "amount_unit",
                    models.CharField(help_text="Unit representation", max_length=25),
                ),
                ("amount_unit_code", models.CharField(max_length=255)),
                (
                    "batch_lot_number",
                    models.CharField(
                        blank=True,
                        help_text="Identifier assigned to batch",
                        max_length=255,
                    ),
                ),
                (
                    "batch_expiration_date",
                    models.DateField(blank=True, help_text="When batch will expire"),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="MedicationForm",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name="MedicationMethod",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                (
                    "maintainer",
                    models.SmallIntegerField(
                        choices=[(1, "User"), (10, "Admin"), (99, "Vendor")], default=1
                    ),
                ),
                ("code", models.CharField(max_length=50)),
                ("display", models.CharField(max_length=50)),
                ("definition", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="MedicationRoute",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                (
                    "maintainer",
                    models.SmallIntegerField(
                        choices=[(1, "User"), (10, "Admin"), (99, "Vendor")], default=1
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="MedicationStatus",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                (
                    "maintainer",
                    models.SmallIntegerField(
                        choices=[(1, "User"), (10, "Admin"), (99, "Vendor")], default=1
                    ),
                ),
                ("code", models.CharField(max_length=50)),
                ("display", models.CharField(max_length=50)),
                ("definition", models.CharField(max_length=255)),
            ],
            options={
                "verbose_name_plural": "Medication Statuses",
            },
        ),
        migrations.CreateModel(
            name="MedicationTerminologySystem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                (
                    "maintainer",
                    models.SmallIntegerField(
                        choices=[(1, "User"), (10, "Admin"), (99, "Vendor")], default=1
                    ),
                ),
                ("name", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Name",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "modified",
                    medux.common.models.AutoDateTimeField(
                        default=django.utils.timezone.now
                    ),
                ),
                ("last_name", models.CharField(max_length=255)),
                ("first_name", models.CharField(max_length=255)),
                (
                    "preferred",
                    models.CharField(
                        blank=True,
                        help_text="preferred first name, the name a person is usually called (nickname)",
                        max_length=255,
                        null=True,
                    ),
                ),
                (
                    "comment",
                    models.CharField(
                        blank=True,
                        help_text="a comment regarding this name, useful in things like 'name before marriage' etc.",
                        max_length=255,
                        null=True,
                    ),
                ),
                ("weight", models.PositiveIntegerField(default=1)),
            ],
            options={
                "ordering": ["weight"],
            },
        ),
        migrations.CreateModel(
            name="Narrative",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "modified",
                    medux.common.models.AutoDateTimeField(
                        default=django.utils.timezone.now
                    ),
                ),
                (
                    "deleted_at",
                    models.DateTimeField(
                        blank=True, default=None, editable=False, null=True
                    ),
                ),
                ("row_version", models.PositiveIntegerField(default=0, editable=False)),
                ("narrative", models.TextField()),
                (
                    "encounter",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="core.encounter"
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="NarrativeType",
            fields=[
                (
                    "shortcut",
                    models.CharField(max_length=2, primary_key=True, serialize=False),
                ),
                ("name", models.CharField(max_length=100)),
            ],
        ),
        migrations.CreateModel(
            name="Observation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "modified",
                    medux.common.models.AutoDateTimeField(
                        default=django.utils.timezone.now
                    ),
                ),
                (
                    "deleted_at",
                    models.DateTimeField(
                        blank=True, default=None, editable=False, null=True
                    ),
                ),
                ("row_version", models.PositiveIntegerField(default=0, editable=False)),
                (
                    "status",
                    medux.core.fields.CodeField(
                        max_length=32,
                        terminology_binding=medux.core.models.observations.Observation.Status,
                    ),
                ),
                (
                    "effective_datetime",
                    models.DateTimeField(
                        blank=True,
                        help_text="Clinically relevant time for observation",
                        null=True,
                    ),
                ),
                ("effective_period_start", models.DateTimeField(blank=True, null=True)),
                ("effective_period_end", models.DateTimeField(blank=True, null=True)),
                ("issued", models.DateTimeField(blank=True, null=True)),
                ("comment", models.CharField(blank=True, max_length=255)),
                (
                    "device",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="core.device",
                    ),
                ),
                (
                    "encounter",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="core.encounter",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Organization",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                (
                    "maintainer",
                    models.SmallIntegerField(
                        choices=[(1, "User"), (10, "Admin"), (99, "Vendor")], default=1
                    ),
                ),
                ("active", models.BooleanField(default=True)),
                ("name", models.CharField(max_length=255)),
                ("alias", medux.core.fields.StringListField(blank=True)),
                (
                    "address",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="core.address",
                    ),
                ),
                ("telecom", models.ManyToManyField(blank=True, to="core.contactpoint")),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="OrganizationType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                (
                    "maintainer",
                    models.SmallIntegerField(
                        choices=[(1, "User"), (10, "Admin"), (99, "Vendor")], default=1
                    ),
                ),
                ("code", models.CharField(max_length=50)),
                ("display", models.CharField(max_length=50)),
                ("definition", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Person",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "modified",
                    medux.common.models.AutoDateTimeField(
                        default=django.utils.timezone.now
                    ),
                ),
                (
                    "deleted_at",
                    models.DateTimeField(
                        blank=True, default=None, editable=False, null=True
                    ),
                ),
                ("row_version", models.PositiveIntegerField(default=0, editable=False)),
                (
                    "title",
                    models.CharField(
                        blank=True, help_text="Academic title", max_length=50, null=True
                    ),
                ),
                (
                    "birth_date",
                    models.DateField(
                        blank=True, null=True, verbose_name="Date of birth"
                    ),
                ),
                (
                    "birth_time",
                    models.TimeField(
                        blank=True, null=True, verbose_name="Time of birth"
                    ),
                ),
                ("birth_date_is_estimated", models.BooleanField(default=False)),
                (
                    "deceased_date",
                    models.DateField(
                        blank=True, null=True, verbose_name="Date of death"
                    ),
                ),
                (
                    "deceased_time",
                    models.TimeField(
                        blank=True, null=True, verbose_name="Time of death"
                    ),
                ),
                ("deceased_is_estimated", models.BooleanField(default=False)),
                ("photo", models.ImageField(blank=True, upload_to="")),
                (
                    "active",
                    models.BooleanField(
                        default=True,
                        help_text="Whether this person's record is in active use.",
                    ),
                ),
                (
                    "addresses",
                    models.ManyToManyField(
                        through="core.AddressMapper", to="core.address"
                    ),
                ),
                (
                    "country_of_birth",
                    models.ForeignKey(
                        help_text="ISO code of Country of Birth",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="core.country",
                    ),
                ),
                (
                    "gender",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        to="core.administrativegender",
                    ),
                ),
                ("names", models.ManyToManyField(to="core.name")),
                (
                    "telecom",
                    models.ManyToManyField(
                        help_text="A contact detail for the person, e.g. a telephone number or an email address",
                        to="core.contactpoint",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Problem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "modified",
                    medux.common.models.AutoDateTimeField(
                        default=django.utils.timezone.now
                    ),
                ),
                (
                    "deleted_at",
                    models.DateTimeField(
                        blank=True, default=None, editable=False, null=True
                    ),
                ),
                ("row_version", models.PositiveIntegerField(default=0, editable=False)),
                ("name", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Relationship",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="RelationType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name="Request",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "created",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False
                    ),
                ),
                (
                    "modified",
                    medux.common.models.AutoDateTimeField(
                        default=django.utils.timezone.now
                    ),
                ),
                (
                    "deleted_at",
                    models.DateTimeField(
                        blank=True, default=None, editable=False, null=True
                    ),
                ),
                ("row_version", models.PositiveIntegerField(default=0, editable=False)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Specialty",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                (
                    "maintainer",
                    models.SmallIntegerField(
                        choices=[(1, "User"), (10, "Admin"), (99, "Vendor")], default=1
                    ),
                ),
                ("name", models.CharField(max_length=255)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="ZipCode",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("uuid", models.UUIDField(default=uuid.uuid4)),
                (
                    "maintainer",
                    models.SmallIntegerField(
                        choices=[(1, "User"), (10, "Admin"), (99, "Vendor")], default=1
                    ),
                ),
                ("code", models.CharField(max_length=10)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="BloodPressure",
            fields=[
                (
                    "observation_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core.observation",
                    ),
                ),
                ("systolic", models.PositiveSmallIntegerField()),
                ("diastolic", models.PositiveSmallIntegerField()),
            ],
            options={
                "abstract": False,
            },
            bases=("core.observation",),
        ),
        migrations.CreateModel(
            name="BodyTemperature",
            fields=[
                (
                    "observation_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core.observation",
                    ),
                ),
                ("value_value", models.DecimalField(decimal_places=1, max_digits=3)),
                ("value_unit", medux.core.fields.UnitField(blank=True, max_length=10)),
                ("value_system", models.CharField(blank=True, max_length=255)),
                (
                    "value_comparator",
                    models.CharField(
                        blank=True,
                        choices=[("<", "<"), ("<=", "<="), (">=", ">="), (">", ">")],
                        max_length=2,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("core.observation",),
        ),
        migrations.CreateModel(
            name="BodyWeight",
            fields=[
                (
                    "observation_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core.observation",
                    ),
                ),
                ("value_value", models.DecimalField(decimal_places=2, max_digits=8)),
                ("value_unit", medux.core.fields.UnitField(blank=True, max_length=10)),
                ("value_system", models.CharField(blank=True, max_length=255)),
                (
                    "value_comparator",
                    models.CharField(
                        blank=True,
                        choices=[("<", "<"), ("<=", "<="), (">=", ">="), (">", ">")],
                        max_length=2,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("core.observation",),
        ),
        migrations.CreateModel(
            name="Company",
            fields=[
                (
                    "organization_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core.organization",
                    ),
                ),
            ],
            options={
                "verbose_name_plural": "Companies",
            },
            bases=("core.organization",),
        ),
        migrations.CreateModel(
            name="ContactPerson",
            fields=[
                (
                    "person_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core.person",
                    ),
                ),
                (
                    "organization",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="contact_persons",
                        to="core.organization",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("core.person",),
        ),
        migrations.CreateModel(
            name="HealthServiceProvider",
            fields=[
                (
                    "organization_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core.organization",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("core.organization",),
        ),
        migrations.CreateModel(
            name="MedicationPlan",
            fields=[
                (
                    "careplan_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core.careplan",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("core.careplan",),
        ),
        migrations.CreateModel(
            name="MedicationRequest",
            fields=[
                (
                    "request_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core.request",
                    ),
                ),
                (
                    "status",
                    medux.core.fields.CodeField(
                        max_length=32,
                        terminology_binding=medux.core.models.medication.MedicationRequest.Status,
                    ),
                ),
                (
                    "intent",
                    medux.core.fields.CodeField(
                        max_length=32,
                        terminology_binding=medux.core.models.medication.MedicationRequest.Intents,
                    ),
                ),
                (
                    "priority",
                    medux.core.fields.CodeField(
                        max_length=32,
                        terminology_binding=medux.core.models.medication.MedicationRequest.Priority,
                    ),
                ),
                (
                    "do_not_perform",
                    models.BooleanField(
                        default=False, help_text="True if request is prohibiting action"
                    ),
                ),
                ("reported", models.BooleanField(default=False)),
                (
                    "note",
                    models.CharField(
                        help_text="Information about the prescription", max_length=255
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("core.request",),
        ),
        migrations.CreateModel(
            name="Patient",
            fields=[
                (
                    "person",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.PROTECT,
                        parent_link=True,
                        primary_key=True,
                        related_name="+",
                        serialize=False,
                        to="core.person",
                    ),
                ),
                (
                    "marital_status",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("A", "Annulled"),
                            ("D", "Divorced"),
                            ("I", "Interlocutory"),
                            ("L", "Legally Separated"),
                            ("M", "Married"),
                            ("P", "Polygamous"),
                            ("S", "Never Married"),
                            ("T", "Domestic partner"),
                            ("U", "Unmarried"),
                            ("W", "Widowed"),
                            ("UNK", "Unknown"),
                        ],
                        max_length=3,
                        null=True,
                    ),
                ),
                (
                    "emergency_contact_freetext",
                    models.CharField(blank=True, max_length=255, null=True),
                ),
                (
                    "comment",
                    models.CharField(
                        blank=True,
                        help_text="Generic comment for this patient, e.g. to identify him/her.",
                        max_length=255,
                        null=True,
                    ),
                ),
                (
                    "emergency_contact",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Persons who should be contacted in case of an emergency",
                        related_name="+",
                        to="core.person",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("core.person",),
        ),
        migrations.DeleteModel(
            name="User",
        ),
        migrations.AddField(
            model_name="relationship",
            name="relation_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="core.relationtype"
            ),
        ),
        migrations.AddField(
            model_name="narrative",
            name="problem",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="core.problem"
            ),
        ),
        migrations.AddField(
            model_name="narrative",
            name="type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="core.narrativetype"
            ),
        ),
        migrations.AddField(
            model_name="medication",
            name="code_system",
            field=models.ForeignKey(
                help_text="The defining coding system",
                on_delete=django.db.models.deletion.PROTECT,
                to="core.medicationterminologysystem",
            ),
        ),
        migrations.AddField(
            model_name="medication",
            name="form",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="core.medicationform"
            ),
        ),
        migrations.AddField(
            model_name="medication",
            name="ingredients",
            field=models.ManyToManyField(to="core.ingredient"),
        ),
        migrations.AddField(
            model_name="episodeofcare",
            name="status",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="core.episodestatus",
            ),
        ),
        migrations.AddField(
            model_name="episodeofcare",
            name="status_history",
            field=models.ManyToManyField(
                through="core.EpisodeStatusHistory", to="core.episodestatus"
            ),
        ),
        migrations.AddField(
            model_name="dosage",
            name="method",
            field=models.ForeignKey(
                blank=True,
                help_text="Technique for administering medication",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="core.medicationmethod",
            ),
        ),
        migrations.AddField(
            model_name="dosage",
            name="route",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="core.medicationroute"
            ),
        ),
        migrations.AddField(
            model_name="device",
            name="owner",
            field=models.ForeignKey(
                help_text="Organization responsible for device",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="core.organization",
            ),
        ),
        migrations.AddField(
            model_name="device",
            name="parent",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="core.device",
            ),
        ),
        migrations.AddField(
            model_name="device",
            name="type",
            field=models.ForeignKey(
                blank=True,
                help_text="The kind or type of device",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="core.devicetype",
            ),
        ),
        migrations.AddField(
            model_name="datapack",
            name="language",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="core.language"
            ),
        ),
        migrations.AddField(
            model_name="country",
            name="default_language",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="core.language",
            ),
        ),
        migrations.AddField(
            model_name="city",
            name="zipcodes",
            field=models.ManyToManyField(to="core.zipcode"),
        ),
        migrations.AddField(
            model_name="careplan",
            name="encounter",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="core.encounter",
            ),
        ),
        migrations.AddField(
            model_name="careplan",
            name="replaces",
            field=models.ForeignKey(
                help_text="CarePlan replaced by this CarePlan",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="replaced_by",
                to="core.careplan",
            ),
        ),
        migrations.AddField(
            model_name="addressmapper",
            name="address_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="core.addresstype"
            ),
        ),
        migrations.AddField(
            model_name="addressmapper",
            name="person_id",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT, to="core.person"
            ),
        ),
        migrations.AddField(
            model_name="address",
            name="state",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="core.country",
            ),
        ),
        migrations.CreateModel(
            name="Hospital",
            fields=[
                (
                    "healthserviceprovider_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core.healthserviceprovider",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("core.healthserviceprovider",),
        ),
        migrations.CreateModel(
            name="InsuranceCompany",
            fields=[
                (
                    "company_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core.company",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("core.company",),
        ),
        migrations.AddField(
            model_name="relationship",
            name="from_patient_id",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="core.patient",
            ),
        ),
        migrations.AddField(
            model_name="relationship",
            name="to_patient_id",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="core.patient",
            ),
        ),
        migrations.AddField(
            model_name="patient",
            name="related_persons",
            field=models.ManyToManyField(
                related_name="+", through="core.Relationship", to="core.patient"
            ),
        ),
        migrations.AddField(
            model_name="observation",
            name="performer",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                to="core.healthserviceprovider",
            ),
        ),
        migrations.AddField(
            model_name="medicationrequest",
            name="based_on",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="core.careplan",
            ),
        ),
        migrations.AddField(
            model_name="medicationrequest",
            name="course_of_therapy_type",
            field=models.ForeignKey(
                blank=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="core.courseoftherapy",
            ),
        ),
        migrations.AddField(
            model_name="medicationrequest",
            name="dosage_instruction",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.PROTECT, to="core.dosage"
            ),
        ),
        migrations.AddField(
            model_name="medicationrequest",
            name="encounter",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="core.encounter"
            ),
        ),
        migrations.AddField(
            model_name="medicationrequest",
            name="medication",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="core.medication"
            ),
        ),
        migrations.AddField(
            model_name="medicationrequest",
            name="recorder",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AddField(
            model_name="medicationrequest",
            name="replaces",
            field=models.ForeignKey(
                help_text="An order/prescription that is being replaced",
                on_delete=django.db.models.deletion.PROTECT,
                related_name="replaced_by",
                to="core.medicationrequest",
            ),
        ),
        migrations.AddField(
            model_name="medicationrequest",
            name="requester",
            field=models.ForeignKey(
                help_text="Who requested the request",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="core.healthserviceprovider",
            ),
        ),
        migrations.AddField(
            model_name="medicationrequest",
            name="subject",
            field=models.ForeignKey(
                help_text="Who medication request is for",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="core.patient",
            ),
        ),
        migrations.AddField(
            model_name="medicationplan",
            name="medications",
            field=models.ManyToManyField(to="core.medicationrequest"),
        ),
        migrations.AddField(
            model_name="medication",
            name="manufacturer",
            field=models.ForeignKey(
                help_text="Manufacturer of the medication product. (Not the distributor.)",
                on_delete=django.db.models.deletion.PROTECT,
                to="core.healthserviceprovider",
            ),
        ),
        migrations.AddField(
            model_name="encounter",
            name="provider",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to="core.healthserviceprovider",
            ),
        ),
        migrations.AddField(
            model_name="device",
            name="patient",
            field=models.ForeignKey(
                blank=True,
                help_text="Patient to whom Device is affixed",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="core.patient",
            ),
        ),
        migrations.AddField(
            model_name="careplan",
            name="author",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="core.healthserviceprovider",
            ),
        ),
        migrations.AddField(
            model_name="careplan",
            name="subject",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="core.patient"
            ),
        ),
        migrations.AddField(
            model_name="bodyweight",
            name="subject",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="core.patient",
            ),
        ),
        migrations.AddField(
            model_name="bodytemperature",
            name="subject",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="core.patient",
            ),
        ),
        migrations.AddField(
            model_name="bloodpressure",
            name="subject",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="core.patient",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="address",
            unique_together={("street", "number", "postcode", "city", "state")},
        ),
        migrations.CreateModel(
            name="Physician",
            fields=[
                (
                    "healthserviceprovider_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="core.healthserviceprovider",
                    ),
                ),
                (
                    "personal_salutation",
                    models.CharField(
                        blank=True,
                        help_text="if you know this physician personally, you can choose to customize his salutation, like 'Dear Thomas,'",
                        max_length=100,
                    ),
                ),
                (
                    "person",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="core.person"
                    ),
                ),
                (
                    "speciality",
                    models.ForeignKey(
                        help_text="Medical field",
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="physicians",
                        to="core.specialty",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("core.healthserviceprovider",),
        ),
        migrations.AddField(
            model_name="patient",
            name="general_practitioner",
            field=models.ForeignKey(
                blank=True,
                help_text="The patient's general practitioner",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="+",
                to="core.physician",
            ),
        ),
        migrations.AddField(
            model_name="patient",
            name="physicians",
            field=models.ManyToManyField(
                blank=True,
                help_text="Physicians the patient consults regularly",
                related_name="+",
                to="core.physician",
            ),
        ),
        migrations.AddField(
            model_name="department",
            name="hospital",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="departments",
                to="core.hospital",
            ),
        ),
    ]



from django.core.exceptions import ValidationError
from django.test import TestCase

from medux.preferences import Scope
from medux.preferences.models import ScopedPreference


class ScopedPreferenceModelTestCase(TestCase):
    def test_create_model(self):
        m = ScopedPreference(
            namespace="foo", key="bar", scope=Scope.VENDOR, value="baz"
        )
        m.save()

        self.assertEqual(m.namespace, "foo")
        self.assertEqual(m.key, "bar")
        self.assertEqual(m.scope, Scope.VENDOR)

    def test_clean_lowercase(self):
        m = ScopedPreference(
            namespace="Foo", key="bAr", scope=Scope.VENDOR, value="baz"
        )
        m.clean()

        self.assertEqual(m.namespace, "foo")
        self.assertEqual(m.key, "bar")

    def test_clean_no_user(self):
        with self.assertRaises(ValidationError):
            m = ScopedPreference(
                namespace="foo", key="bar", scope=Scope.USER, value="baz"
            )
            m.clean()

    def test_clean_no_group(self):
        with self.assertRaises(ValidationError):
            m = ScopedPreference(
                namespace="foo", key="bar", scope=Scope.GROUP, value="baz"
            )
            m.clean()

    # def test_clean_no_device(self):
    #     with self.assertRises(ValidationError):
    #         m = ScopedPreference(
    #             namespace="foo", key="bar", scope=Scope.DEVICE, value="baz"
    #         )
    #         m.clean()

    def test_clean_no_tenant(self):
        with self.assertRaises(ValidationError):
            m = ScopedPreference(
                namespace="foo", key="bar", scope=Scope.TENANT, value="baz"
            )
            m.clean()



from .api.interfaces import IMenuItem


class Menu:
    """Represents a named menu during a request.

    Usually it is used in Django templates by calling ``menus.<name>``
    which then will return all ``MenuItem``s with a matching "menu" name
    attribute. Therefore, Menu will be instantiated by a context_processor,
    so that the ``menu`` variable in the template

    .. code-block:: django

        <ul>
        {% for item in menus.user %}
            <li><a href="{{item.url}}">{{ item.title }}</a></li>
        {% endfor %}
        </ul>
    """

    def __init__(self, request):
        self.request = request
        self._cache = []
        for menu_item_class in IMenuItem:
            item = menu_item_class(self.request)
            self._cache.append(item)

    def _items(self):
        """walk through"""

    def __getitem__(self, item):
        """returns filtered out menu items with the given '.menu' name."""
        for menu_item in self._cache:
            # TODO: handle submenus
            if (
                menu_item.menu == item  # only items in requested menu
                and "." not in menu_item.slug  # only top level items
                and menu_item.visible  # only visible items
            ):
                yield menu_item



import logging
from django.contrib.auth.signals import (
    user_logged_in,
    user_logged_out,
    user_login_failed,
)

from django.dispatch import receiver

logger = logging.getLogger(__name__)

# https://stackoverflow.com/questions/37618473/how-can-i-log-both-successful-and-failed-login-and-logout-attempts-in-django
@receiver(user_logged_in)
def user_logged_in_callback(sender, request, user, **kwargs):
    logger.debug(f"login user: {user}")


@receiver(user_logged_out)
def user_logged_out_callback(sender, request, user, **kwargs):
    ip = request.META.get("REMOTE_ADDR")
    logger.debug(f"logout user: {user} via ip: {ip}")


@receiver(user_login_failed)
def user_login_failed_callback(sender, credentials, **kwargs):
    logger.warning(f"login failed for: {credentials}")



from django.utils.translation import gettext_lazy as _
from medux.common.api import MeduxPluginAppConfig
from . import __version__


class AlertsConfig(MeduxPluginAppConfig):
    """A GDAPS Django app plugin."""

    # Dotted Python path to the application, e.g. '/home/christian/Projekte/medux/medux/plugins.fooplugin'
    # It has to start with '/home/christian/Projekte/medux/medux/plugins' to be recognized as plugin.
    name = "medux.alerts"

    # This is the most important attribute of a GDAPS plugin app.
    class PluginMeta:
        """This configuration is the introspection data for plugins."""

        # the plugin machine "name" is taken from the AppConfig, so no name here
        verbose_name = _("Alerts")
        author = "Christian González"
        author_email = "christian.gonzalez@nerdocs.at"
        vendor = "Nerdocs"
        description = _("MedUX Alerts plugin")
        category = _("Core")
        visible = True
        version = __version__
        # compatibility = "medux.core>=2.3.0"

    def ready(self):
        try:
            from . import signals
        except ImportError:
            pass

        # import implementations of global interfaces
        from . import impl



from django import template
from django.db.models import Model

register = template.Library()


@register.filter
def class_name(model: Model):
    """:returns the singular verbose name of the given object's class.

    This can be used in templates:

    .. code-block:: django

        <div>Should this {{ object|class_name }} really be deleted?</div>
    """
    return model._meta.verbose_name


@register.filter
def class_name_plural(model: Model):
    """:returns the plural verbose name of the given object's class.

    This can be used in templates:

    .. code-block:: django

        <div>Should this {{ object|class_name_plural }} really be deleted?</div>
    """
    return model._meta.verbose_name_plural



"""
MedUX - A Free/OpenSource Electronic Medical Record
Copyright (C) 2017-2022 Christian González

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

# TODO: Fields could be done better, see http://build.fhir.org/datatypes.html
# These types should be implemented very carefully, with all the validators
# and custom behaviours in place.
__author__ = "Christian González <christian.gonzalez@nerdocs.at>"

import base64
from datetime import timedelta
from decimal import Decimal
from enum import Enum
from uuid import uuid4

from composite_field import CompositeField
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.validators import RegexValidator, MinValueValidator
from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from enumfields import EnumField

from .validators import CodeValidator, OidValidator, IdValidator


class OptionalTimeDateTimeField(models.DateTimeField):
    """Field that optionally uses the time format.

    Set MEDUX["BIRTHDAY_FORMAT"] in settings.py to either "Date" or "DateTime" to use the desired format.
    """

    def formfield(self, **kwargs):
        field = None
        try:
            if settings.MEDUX["BIRTHDAY_FORMAT"] == "Date":
                field = "forms.DateField"
            elif settings.MEDUX["BIRTHDAY_FORMAT"] == "DateTime":
                field = "forms.DateTimeField"
            else:
                raise ImproperlyConfigured(
                    f"MEDUX['BIRTHDAY_FORMAT'] settings variable contains wrong value '{settings.MEDUX['BIRTHDAY_FORMAT']}'. Please set it to 'Date' or 'DateTime'."
                )
        except AttributeError:
            raise ImproperlyConfigured(
                "MEDUX settings variable does not contain a 'BIRTHDAY_FORMAT'. Please set it to 'Date' or 'DateTime'."
            )
        # FIXME: DateField is ignored in Admin
        return super().formfield(
            **{
                "form_class": field,
                **kwargs,
            }
        )


class Base64TextField(models.TextField):
    """A stream of bytes, base64 encoded"""

    # TODO: +RegexValidator '(\s*([0-9a-zA-Z\+\=]){4}\s*)+'
    def from_db_value(value):
        """Returns a str from the database value,
        which is encoded as base64 string"""

        if value is None:
            return None
        else:
            return base64.b64decode(value).decode("utf-8")

    # TODO: read only yet


class StringListField(models.TextField):
    """Represents a list of strings.

    in FHIR resources, there is often a string[0..*] needed. As Django
    Doesn't provide a string list field, this is a simple implementation.
    Other implementations would be possible using JSONfield on PostGreSQL."""

    # TODO: implement this field completely.
    # TODO: Beware of strings >255 chars
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    # def to_python(self, value):
    #     """Normalize data to a list of strings."""
    #     # Return an empty list if no input was given.
    #     if not value:
    #         return []
    #     return value.split("\n")


class InstantField(models.DateTimeField):
    """An instant in time

    It must be known at least to the second and always includes a time zone.
    Note: This type is for system times, not human times."""

    def __init__(self, *args, **kwargs):
        kwargs["validators"] = [
            RegexValidator(
                r"([0-9]([0-9]([0-9][1-9]|[1-9]0)|"
                r"[1-9]00)|[1-9]000)-(0[1-9]|1[0-2])-"
                r"(0[1-9]|[1-2][0-9]|3[0-1])T([01][0-9]|2[0-3]):"
                r"[0-5][0-9]:"
                r"([0-5][0-9]|60)(\.[0-9]+)?(Z|(\+|-)"
                r"((0[0-9]|1[0-3]):[0-5][0-9]|14:00))"
            )
        ]
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()

        del kwargs["validators"]
        return name, path, args, kwargs


class UriField(models.URLField):
    """A Uniform Resource Identifier Reference.

    This is a URI, as defined in RFC 3986: https://tools.ietf.org/html/rfc3986
    Note: URIs generally are case sensitive. For an UUID like
    (urn:uuid:ad1b1c1b-96b0-4c4d-a826-2b6e31f0512b) use all lowercase letters!
    """

    # TODO: implementation
    def __init__(self, *args, **kwargs):
        # FIXME: we set this to an arbitrary 255 char string as max. could be more specific
        kwargs["max_length"] = 255
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["max_length"]
        return name, path, args, kwargs


class UrlField(UriField):
    """A Uniform Resource Locator (RFC 1738 ).

    Note: URLs are accessed directly using the specified protocol.
    Common URL protocols are http(s):, ftp:, mailto: and mllp:,
    though many others are defined.
    """


class CanonicalField(UriField):
    """A URI that refers to a canonical URI.

    The canonical type differs from a uri in that it has special meaning
    in this specification, and in that it may have a version appended,
    separated y a vertical bar (|).

    URIs can be absolute or relative, and may have an optional fragment
    identifier.
    This data type can be bound to a value set
    """


# FIXME: better use TextChoices here?
class CodeField(EnumField):
    """Represents a field with a "code" that is defined elsewhere.

    This field must not contain arbitrary strings, but only the ones that are defined in the given
    terminology system.

    Technically, a code is restricted to a string which has at least one
    character and no leading or trailing whitespace, and where there is no
    whitespace other than single spaces in the contents.

    A code is determined by 4 elements:

    * system - the system the code lives in, e.g. "http://loinc.org"
    * version - the version this code has in the system, e.g. "1.27"
    * code - the code itself, e.g. "55423-8", or "active"
    * display - the human readable representation, e.g. "The prescription is active"
    """

    # TODO: This could be a ForeignKey, pointing to a FHIR ValueSet class!
    description = _('A "code" that is defined elsewhere')
    default_validators = [CodeValidator]
    # default_error_messages = {
    #     "not_multiple": _('There are no multiple values allowed in "{key}".'),
    # }

    def __init__(
        self,
        terminology_binding: type(Enum) = None,
        # system: str = None,
        # version: str = None,
        **kwargs,
    ):

        if terminology_binding is not None and not issubclass(
            terminology_binding, Enum
        ):
            raise TypeError(
                f"CodeField: 'technology_binding' parameter must be a Enum, but is '{type(terminology_binding)}'"
            )

        # TODO: do something with system/version etc.
        # self.system = system
        # self.version = version
        # The ValueSet this code is defined in
        self.terminology_binding = terminology_binding

        # if not value_set:
        #     raise exceptions.ValidationError(
        #         "Terminology binding {} not found.".format(terminology_binding)
        #     )
        # if terminology_binding:
        #     kwargs["choices"] = [(tag.value, tag.name) for tag in terminology_binding]
        # if not terminology_binding:
        #     terminology_binding = Enum
        kwargs["max_length"] = 32
        if "enum" in kwargs:
            print("found!", kwargs)
        super().__init__(terminology_binding, **kwargs)

    def validate(self, value, model_instance):
        super().validate(value, model_instance)
        # print(value)
        #        for key, val in value.items():
        #            if True:
        #                raise exceptions.ValidationError(
        #                    self.error_messages['not_multiple'],
        #                    code='not_multiple',
        #                    params={'key': key},
        #                )
        return

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["enum"]
        kwargs["terminology_binding"] = self.terminology_binding
        return name, path, args, kwargs


class OidField(UriField):
    """An OID represented as a URI"""

    def __init__(self, *args, **kwargs):
        kwargs["validators"] = [OidValidator]
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["validators"]
        return name, path, args, kwargs


class IdField(models.CharField):
    """A field that can be used for an ID of an Object.

    Any combination of upper or lower case ASCII letters ('A'..'Z', and 'a'..'z',
    numerals ('0'..'9'), '-' and '.', with a length limit of 64 characters.
    This might be an integer, an un-prefixed OID, UUID or any other identifier
    pattern that meets these constraints."""

    def __init__(self, *args, **kwargs):
        kwargs["max_length"] = 64
        kwargs["default"] = uuid4
        kwargs["validators"] = [IdValidator]
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        del kwargs["max_length"]
        del kwargs["default"]
        del kwargs["validators"]
        return name, path, args, kwargs


class MarkdownField(models.TextField):
    """A string that *may* contain markdown syntax.

    This can be used for optional processing by a markdown presentation engine"""

    # TODO: implement Validator: \s*(\S|\s)*


class NarrativeStatus(Enum):
    GENERATED = "generated"
    EXTENSIONS = "extensions"
    ADDITIONAL = "additional"
    EMPTY = "empty"


class PeriodField(CompositeField):
    description = "A Field representing a period in time, with optional start or end."

    # Constraints: One of [start|end] must be given.

    start = models.DateTimeField(blank=True, null=True)
    end = models.DateTimeField(blank=True, null=True)

    # def __repr__(self):
    #     return Period(start=self.start, end=self.end)

    def timedelta(self) -> timedelta:
        return self.end - self.start


class UnitField(models.CharField):
    description = "A field which stores units."

    def __init__(self, **kwargs):
        kwargs.setdefault("max_length", 10)
        super().__init__(**kwargs)


class QuantityField(CompositeField):
    COMPARATORS = (
        ("<", _("<")),
        ("<=", _("<=")),
        (">=", _(">=")),
        (">", _(">")),
    )

    description = (
        "A field which stores value and (optionally) unit of a quantity., e.g. '250 ml'"
    )

    value = models.DecimalField(max_length=None, decimal_places=None)
    unit = UnitField(blank=True)
    system = models.CharField(max_length=255, blank=True)
    # code = CodeField(terminology_binding=...)  # The (FHIR) preferred system is UCUM
    comparator = models.CharField(max_length=2, blank=True, choices=COMPARATORS)

    def __init__(
        self,
        max_digits: int | None,
        decimal_places: int | None,
        verbose_name: str = None,
        blank: bool = False,
        default: Decimal = None,
        validators=None,
    ):
        super().__init__()
        self.verbose_name = verbose_name
        value = self["value"]
        value.blank = blank
        value.decimal_places = decimal_places
        value.max_digits = max_digits
        value.null = blank
        value.blank = blank
        self._validators = validators
        if default is not None:
            self.value.default = default.real

    def __str__(self):
        return f"{self.value} {self.unit}"

    # def clean(self, value, model_instance):
    #     """
    #     We need to run validation against ``Quantity`` instance.
    #     """
    #     output = self.to_python(value)
    #     self.validate(value, model_instance)
    #     self.run_validators(value)
    #     return output

    @cached_property
    def validators(self):
        # Default ``DecimalValidator`` doesn't work with ``Quantity`` instances.
        # FIXME use correct validators
        return super().validators + [self._validators]

    def formfield(self, **kwargs):
        defaults = {
            "form_class": "core.QuantityFormField",
            "decimal_places": self.value.decimal_places,
        }
        defaults.update(kwargs)
        return super().formfield(**defaults)


class RatioField(CompositeField):
    """A relationship between two Quantity values.

    Examples where a Ratio is used are:

    * titers (e.g. 1:150)
    * concentration ratios where the denominator is significant (e.g. 5mg/dl)
    * observed frequencies (e.g. 2 repetitions/8 hr)
    """

    numerator = QuantityField(
        max_digits=1, decimal_places=1, blank=True
    )  # 1=placeholder
    denominator = QuantityField(
        max_digits=1, decimal_places=1, validators=[MinValueValidator(0)]
    )  # 1=placeholder

    def __init__(
        self,
        num_max_digits: int,
        num_decimal_places: int,
        denom_max_digits: int,
        denom_decimal_places: int,
        verbose_name: str = None,
        blank: bool = False,
        validators=None,
    ):
        super().__init__()
        self.verbose_name = verbose_name

        numerator = self["numerator"]
        numerator.max_digits = num_max_digits
        numerator.decimal_places = num_decimal_places

        denominator = self["denominator"]
        denominator.max_digits = denom_max_digits
        denominator.decimal_places = denom_decimal_places

        numerator.blank = blank
        denominator.blank = blank
        denominator.null = False  # TODO: make sure null works here
        self._validators = validators

    def __str__(self):
        return f"{self.numerator}/{self.denominator}"


# ======================== Unused Fields from FHIR ========================


# ===================== Primitive types =====================
# Primitive types are those with only a value,
# and no additional elements as children.
# see examples: http://build.fhir.org/datatypes-examples.html#primitives
# ===========================================================


# StringFields are implemented either as Textfields, or as Charfields,
# depending on how many chars (>255) are possible. Some databases do not
# implement more than 255 chars in a string.


# ===================== Complex types =====================
# In XML, these types are represented as XML Elements with child elements
# with the name of the defined elements of the type.
# The name of the element is defined where the type is used.
# In JSON, the data type is represented by an object with properties named
# the same as the XML elements. Since the JSON representation is almost
# exactly the same, only the first example has an additional explicit
# JSON representation.
#
# Complex data types may be "profiled".
# A Structure Definition or type "constraint" makes a set of rules about
# which elements SHALL have values and what the possible values are.
# ===========================================================


# class ManyReferenceField(models.ManyToManyField):
#     def __init__(self, to: str, **kwargs):
#
#         self.allowed_references = to.split("|")
#         for ref in self.allowed_references:
#             if ref not in fhir_server_allowed_references.split("|"):
#                 raise exceptions.FieldError(
#                     _("'{}' is not allowed as reference in {}".format(ref, "<FIXME>"))
#                 )
#
#         # No matter what this field should (dynamically) refer to,
#         # always make sure the Foreign key is bound to a "Reference" object
#         kwargs["to"] = "Reference"
#         super().__init__(**kwargs)
#
#     def deconstruct(self):
#         name, path, args, kwargs = super().deconstruct()
#         kwargs["to"] = "|".join(self.allowed_references)
#         return name, path, args, kwargs


# class CodeableConceptField(models.ForeignKey):
#     def __init__(self, value_set: str, *args, **kwargs):
#         assert type(value_set) == str
#
#         self.allowed_codings = value_set.split("|")
#
#         # always refer to the model "CodeableConcept", no matter what is given
#         kwargs["to"] = "CodeableConcept"
#
#         # FIXME: is SET_NULL ok everywhere?
#         kwargs["null"] = True
#         kwargs["on_delete"] = models.SET_NULL
#         super().__init__(*args, **kwargs)
#
#     def deconstruct(self):
#         name, path, args, kwargs = super().deconstruct()
#         del kwargs["null"]
#         kwargs["value_set"] = "|".join(self.allowed_codings)
#         return name, path, args, kwargs


# class PositiveIntField(models.PositiveIntegerField):
#     description = _("Positive integer >= 1")
#
#     def formfield(self, **kwargs):
#         defaults = {"min_value": 1}
#         defaults.update(kwargs)
#         return super().formfield(**defaults)


# class UnsignedIntField(models.PositiveIntegerField):
#     description = _("Positive integer >= 0")
#
#     # this is the Django implementation of PositiveIntegerField


# http://hl7.org/fhir/narrative-status
# FIXME: implement/import as ValueSet


# ======================== Special data types ========================


# class ReferenceField(models.ForeignKey):
#     """A field that holds a Foreignkey to a Reference Object,
#     which points to another FHIR Resource
#
#     At least one of reference, identifier and display SHALL be present
#     (unless an extension is provided).
#     The Reference object should return the real object.
#     FIXME this has to be coded in Django, does not work yet.
#     """
#
#     # This regex is true if the reference to a resource is consistent with a FHIR API
#     fhir_server_abs_url_conformance = (
#         r"((http|https)://([A-Za-z0-9\\\.\:\%\$]\/)*)?("
#         + fhir_server_allowed_references
#         + ")\/[A-Za-z0-9\-\.]{1,64}(\/_history\/[A-Za-z0-9\-\.]{1,64})?"
#     )
#
#     description = _("A dynamic reference to another Resource")
#
#     def __init__(self, to: str, **kwargs):
#         """Creates a new Reference Field.
#
#         The 'to' parameter is always overwritten and set to "Reference",
#         as the Reference db table works as intermediate mapper to the "real"
#         reference where this field points to.
#
#         Note: the 'on_delete' parameter must be set manually, as it could
#         change according to the context.
#         :param str to: one or more possible FHIR resources where an object
#             could point to. If there are more than one, use '|' as separator.
#             Example: 'Patient|Practitioner|Organization'
#         :raises FieldError: if the 'to' parameter is a not allowed resource.
#         """MarkdownField
#
#         assert type(to) == str
#
#         self.allowed_references = to.split("|")
#         for ref in self.allowed_references:
#             if ref not in fhir_server_allowed_references.split("|"):
#                 raise exceptions.FieldError(
#                     _("'{}' is not allowed as reference in {}".format(ref, "<FIXME>"))
#                 )
#
#         # No matter what this field should (dynamically) refer to,
#         # always make sure the Foreign key is bound to a "Reference" object
#         kwargs["to"] = "Reference"
#         super().__init__(**kwargs)
#
#     def deconstruct(self):
#         name, path, args, kwargs = super().deconstruct()
#         kwargs["to"] = "|".join(self.allowed_references)
#         return name, path, args, kwargs


# class NarrativeField(models.TextField):
#
#     # http://hl7.org/fhir/ValueSet/narrative-status
#     # general, extensions, additional, empty
#     status = models.CharField(max_length=35, choices=NARRATIVE_STATUS)
#
#     # TODO: implement a XHTMLField
#     # The XHTML content SHALL NOT contain a head, a body element, external stylesheet references,
#     # deprecated elements, scripts, forms, base/link/xlink, frames, iframes, objects or event related attributes
#     # (e.g. onClick).This is to ensure that the content of the narrative is contained within the resource
#     # and that there is no active content. Such content would introduce security issues and potentially safety
#     # issues with regard to extracting text from the XHTML.
#     div = models.TextField(null=False, blank=False)


# def validate_quantity_value(value):
#     """
#     Valid value for Quantity are:
#       - Single numeric value
#       - Quantity instances
#       - Pairs of numeric value and unit. unit can't be None.
#     """
#     if isinstance(value, (list, tuple)) and (len(value) != 2 or value[1] is None):
#         raise ValidationError(
#             f"Invalid value for QuantityField: {value}.",
#             code="invalid",
#         )
#



from django_unicorn.components import UnicornView


class MedicationRequestLine(UnicornView):
    # template_name = "core/medicationrequest_line.html"

    def __init__(self, *args, **kwargs):
        super().__init__(**kwargs)
        # self.name = kwargs.get("name")

    def search(self, **kwargs):
        med_text = kwargs["medication_name"]
        self.med_text = med_text



import logging
import os
from pathlib import Path

import requests
from django.apps import apps
from django.core.management import BaseCommand

logger = logging.getLogger(__file__)


class Command(BaseCommand):
    """Download libraries and save them into the static folder.

    The `libradies` dict has the name of the library as key, and a list of
    3-tuples, each containing:
      - the local file path of the downloaded file
      - the URL to download the file
      - a mode how to write the file: "w" for text files, "wb" for binary files like icon fonts

    """

    libraries: dict[list[tuple[str, str, str]]] = {
        "htmx": [
            ("js/htmx/htmx.js", "https://unpkg.com/htmx.org@latest/dist/htmx.js", "w"),
            (
                "js/htmx/htmx.min.js",
                "https://unpkg.com/htmx.org@latest/dist/htmx.min.js",
                "w",
            ),
            (
                "js/htmx/ext/ws.js",
                "https://unpkg.com/htmx.org@latest/dist/ext/ws.js",
                "w",
            ),
            (
                "js/htmx/ext/debug.js",
                "https://unpkg.com/htmx.org@latest/dist/ext/debug.js",
                "w",
            ),
        ],
        # "bootstrap": [
        #     (
        #         "js/bootstrap.bundle.min.js",
        #         "https://unpkg.com/bootstrap@latest/dist/js/bootstrap.bundle.min.js",
        #         "w",
        #     ),
        #     (
        #         "js/bootstrap.bundle.min.js.map",
        #         "https://unpkg.com/bootstrap@latest/dist/js/bootstrap.bundle.min.js.map",
        #         "w",
        #     ),
        #     (
        #         "css/bootstrap.min.css",
        #         "https://unpkg.com/bootstrap@latest/dist/css/bootstrap.min.css",
        #         "w",
        #     ),
        #     (
        #         "css/bootstrap.min.css.map",
        #         "https://unpkg.com/bootstrap@latest/dist/css/bootstrap.min.css.map",
        #         "w",
        #     ),
        # ],
        "bootstrap-icons": [
            (
                "css/bootstrap-icons.css",
                "https://unpkg.com/bootstrap-icons@latest/font/bootstrap-icons.css",
                "w",
            ),
            (
                "css/fonts/bootstrap-icons.woff",
                "https://unpkg.com/bootstrap-icons@latest/font/fonts/bootstrap-icons.woff",
                "wb",
            ),
            (
                "css/fonts/bootstrap-icons.woff2",
                "https://unpkg.com/bootstrap-icons@latest/font/fonts/bootstrap-icons.woff2",
                "wb",
            ),
        ],
        "sortable": [
            (
                "js/Sortable.min.js",
                "https://cdn.jsdelivr.net/npm/sortablejs@latest/Sortable.min.js",
                "w",
            )
        ],
        "tabler": [
            (
                "css/tabler.css",
                "https://raw.githubusercontent.com/nerdoc/tabler/medux/dist/css/tabler.css",
                "w",
            ),
            (
                "js/tabler.js",
                "https://raw.githubusercontent.com/nerdoc/tabler/medux/dist/js/tabler.js",
                "w",
            ),
        ],
    }
    help = "DEV COMMAND: downloads needed MedUX Javascript/CSS libraries and fonts."

    def add_arguments(self, parser):
        parser.add_argument(
            "library",
            help=f"Specify the library do download. Available are: {[l for l in self.libraries]}",
        )

    def handle(self, *args, **options):
        target_path_root = (
            Path(apps.get_app_config("common").path) / "static" / "common"
        )
        if not options["library"] in self.libraries:
            return f"No library named '{options['library']}' found. Available are: {[l for l in self.libraries]}"

        library = self.libraries[options["library"]]
        for file_name, url, mode in library:
            data = requests.get(url)
            os.makedirs(os.path.dirname(target_path_root / file_name), exist_ok=True)
            self.stdout.write(file_name)
            with open(target_path_root / file_name, mode) as f:
                if mode == "w":
                    f.write(data.text)
                elif mode == "wb":
                    f.write(data.content)



from crispy_bootstrap5.bootstrap5 import FloatingField
from crispy_forms.bootstrap import InlineField
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, ButtonHolder, Submit
from django import forms
from django.contrib.auth.forms import (
    AuthenticationForm as DjangoAuthenticationForm,
    UserCreationForm as DjangoUserCreationForm,
)
from django.utils.translation import gettext_lazy as _

from medux.common.bootstrap import Card
from medux.common.models import User


class AuthenticationForm(DjangoAuthenticationForm):

    remember_me = forms.BooleanField(
        label=_("Remember me"), widget=forms.CheckboxInput(), required=False
    )

    class Meta:
        model = User
        fields = ("username", "password")

    def __init__(self, request, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.fields["username"].help_text = ""
        self.fields["username"].widget.attrs.update({"autofocus": True})
        self.fields["password"].help_text = ""

        self.helper = FormHelper()
        self.helper.layout = Layout(
            Card(
                FloatingField("username"),
                FloatingField("password"),
                InlineField("remember_me"),
                title=_("Please sign in to get access to your site."),
            ),
            ButtonHolder(Submit("submit", _("Sign in"), css_class="w-100 btn-lg")),
        )


class SignUpForm(DjangoUserCreationForm):
    username = forms.CharField(
        widget=forms.TextInput(
            attrs={"placeholder": "Username", "class": "form-control"}
        )
    )
    email = forms.EmailField(
        widget=forms.EmailInput(attrs={"placeholder": "Email", "class": "form-control"})
    )
    password1 = forms.CharField(
        widget=forms.PasswordInput(
            attrs={"placeholder": "Password", "class": "form-control"}
        )
    )
    password2 = forms.CharField(
        widget=forms.PasswordInput(
            attrs={"placeholder": "Password check", "class": "form-control"}
        )
    )

    class Meta:
        model = User
        fields = ("username", "email", "password1", "password2")



from django.contrib import messages

LEVEL_ICONS = {
    messages.DEBUG: "bug-fill",
    messages.INFO: "info-circle-fill",
    messages.SUCCESS: "check2-circle",
    messages.WARNING: "exclamation-triangle-fill",
    messages.ERROR: "exclamation-square-fill",
}

USERS_GROUP_NAME = "Users"



from django.contrib.auth import get_user_model

from .constants import USERS_GROUP_NAME

User = get_user_model()


def add_user_to_users_group(sender, instance, **kwargs):
    """Signal receiver that is called after each creation of a User
    and adds this user to the "Users" group.
    """
    from django.contrib.auth.models import Group

    if not sender == User:
        return

    users_group, created = Group.objects.get_or_create(name=USERS_GROUP_NAME)
    if created:
        users_group.save()

    # add "Users" group to user's groups
    instance.groups.add(users_group)



import logging

from django.http.response import HttpResponseNotAllowed

logger = logging.getLogger(__file__)


class HtmxResponseMixin:
    """View Mixin to add HTMX functionality.

    If enforce_htmx is True, all requests that do not come from a HTMX
    component are blocked.

    When a HTMX request is detected, the view adds ``_htmx`` to the
    template name. So you can e.g. use different templates for normal
    requests and htmx requests (e.g.  ``model_list.html`` vs
     ``model_list_htmx.html``).
    You can always override that behaviour by explicitly specifying
    a fixed ``template_name`` attribute.
    """

    enforce_htmx = False

    def get_template_names(self):
        # if self.request.htmx:
        #     if self.template_name_suffix:
        #         self.template_name_suffix += "_htmx"
        #     else:
        #         self.template_name_suffix = "_htmx"
        return super().get_template_names()

    def dispatch(self, request, *args, **kwargs):
        if self.enforce_htmx and not self.request.htmx:
            return HttpResponseNotAllowed(
                f"Permission denied: View {self.__class__.__name__} can only be called "
                f"by HTMX."
            )
        response = super().dispatch(request, *args, **kwargs)
        # logger.debug(f"View {self.__class__.__name__} called from HTMX.")
        # FIXME: this should be placed in MedUX' middleware
        # if messages.get_messages(request):
        #     response.headers["HX-Trigger"] = "showMessages"
        return response

    # def get_context_data(self, **kwargs):
    #     context = super().get_context_data(**kwargs)
    #     if self.request.htmx:
    #         context["base_template"] = self.base_template_htmx
    #         print(
    #             f"{self.__class__.__name__}: using {self.base_template_htmx} as base_template."
    #         )
    #     else:
    #         context["base_template"] = self.base_template
    #         print(
    #             f"{self.__class__.__name__}: using {self.base_template} as base_template."
    #         )
    #     return context



import logging

from medux.preferences.definitions import KeyType, Scope

logger = logging.getLogger(__file__)


class PreferencesRegistry:
    """A in-memory storage where preferences namespaces/keys/scopes are registered at
    application start and can be checked against for existence and type.

    Every available setting *must* be registered before usage.

    You can add a (translatable) help_text and an icon to each key here,
    which can be retrieved from the setting itself later.
    """

    # namespace, key, scope
    _registered_keys: list[tuple[str, str, Scope, str]] = []
    _key_types: dict[tuple[str, str], KeyType] = {}
    _help_texts: dict[tuple[str, str], str] = {}
    _icons: dict[tuple[str, str], str] = {}

    def __init__(self):
        raise NotImplementedError(
            f"{self.__class__.__name__} is not meant to be instantiated."
        )

    @classmethod
    def register(
        cls,
        namespace: str,
        key: str,
        allowed_scopes: list["Scope"] = None,
        key_type: KeyType = KeyType.STRING,
        help_text: str = "",
        icon: str = "gear",
    ):
        """Registers a preferences variable in a global list.

        This makes sure that preferences variables read in from e.g. .toml files can only be accepted if they
        match existing, already registered preferences.
        :param namespace: The namespace of the key
        :param key: the key name
        :param allowed_scopes: a list of allowed `Scope`s. VENDOR will be added automatically.
        :param key_type: the type of the setting. Other types are not allowed at assignment later.
        :param help_text: the (translatable) help text that explains in short words what this setting is for
        :param icon: the Bootstrap5 icon used for this setting
        """

        # VENDOR scope is always allowed
        if allowed_scopes is None:
            allowed_scopes = []
        if Scope.VENDOR not in allowed_scopes:
            allowed_scopes.append(Scope.VENDOR)

        cls._help_texts[(namespace, key)] = help_text
        for scope in allowed_scopes:
            t = (namespace, key, scope)
            # can't register same key again...
            if t in cls._registered_keys:
                logger.warning(
                    f"Key {namespace}.{key} [{scope.name}] was already registered!"
                )
            else:
                cls._registered_keys.append(t)
                cls._key_types[(namespace, key)] = key_type
                cls._icons[(namespace, key)] = icon

    @classmethod
    def exists(cls, namespace: str, key: str, scope: Scope) -> bool:
        """:return: True if there was a setting with given namespace.key/scope registered."""

        # return (namespace, key, scope) in cls._registered_keys
        return (namespace, key, scope) in cls._registered_keys

    @classmethod
    def scopes(cls, namespace: str, key: str) -> set[Scope]:
        """:return: a set of scopes registered under the given namespaced preferences key."""
        scopes = set()
        for s in cls._registered_keys:
            if s[0] == namespace and s[1] == key:
                scopes.add(s[2])
        return scopes

    @classmethod
    def key_type(cls, namespace: str, key: str) -> KeyType:
        """:return: the type of the given namespaced preferences key."""
        t = (namespace, key)
        if t in cls._key_types:
            return cls._key_types[t]
        else:
            raise KeyError(f"No setting '{namespace}.{key}' registered.")

    @classmethod
    def all_dct(cls) -> dict[str, dict[str, str]]:
        _all = {}
        for namespace, key, scope in cls._registered_keys:
            if namespace not in _all:
                _all[namespace] = {}
            if key not in _all[namespace]:
                _all[namespace][key] = {}

            _all[namespace][key][scope] = ""
        return _all

    @classmethod
    def all(cls) -> tuple[str, str, str, str]:
        """:return: A tuple of all registered preferences keys: namespace, key, scope, key_type"""
        for setting_tuple in cls._registered_keys:
            yield setting_tuple + (cls._key_types[setting_tuple[0:2]],)

    @classmethod
    def help_text(cls, namespace: str, key: str) -> str:
        return cls._help_texts[(namespace, key)]

    @classmethod
    def icon(cls, namespace: str, key: str) -> str:
        """:return: The Boostrap5 icon name of this setting."""
        return cls._icons[(namespace, key)]

    def delete_orphaned_preferences():
        """Finds orphaned keys and deletes them.

        Orphaned preferences whose Registry key was deleted during
        evolvement of MedUX still stay in the database and may produce
        problems.
        This procedure deletes every setting that is not registered
         in the PreferencesRegistry any more.
        """
        from .registry import PreferencesRegistry
        from .models import ScopedPreference

        orphaned_ids = []
        for item in ScopedPreference.objects.all():
            if not PreferencesRegistry.exists(item.namespace, item.key, item.scope):
                # FIXME: take care of "orphaned" keys which are bound to a user, device, or tenant.
                orphaned_ids.append(item.id)
        ScopedPreference.objects.filter(id__in=orphaned_ids).delete()



from django.db.models import PositiveIntegerField
from django.utils.translation import gettext_lazy as _
from django.db import models

import medux.core.fields
from medux.common.models import CreatedModifiedModel, BaseModel
from .fhir import CodeValueSet, PackageDataModel, ContactPoint, AdministrativeGender
from .geo import Country, Address


class Specialty(PackageDataModel):
    """Medical field a physician operates in"""

    name = models.CharField(max_length=255)

    def __str__(self):
        return self.name


# https://terminology.hl7.org/CodeSystem/organization-type
class OrganizationType(CodeValueSet):
    pass


class Organization(PackageDataModel):
    active = models.BooleanField(default=True)
    name = models.CharField(max_length=255)
    alias = medux.core.fields.StringListField(blank=True)

    # FIXME: should not be mandatory!
    # phone, fax, email etc.
    telecom = models.ManyToManyField(ContactPoint, blank=True)
    address = models.ForeignKey(
        Address, null=True, blank=True, on_delete=models.PROTECT, related_name="+"
    )

    def __str__(self):
        return self.name

    @property
    def type(self):
        """Implement this type for each inheriting class.

        See https://terminology.hl7.org/CodeSystem/organization-type for a list."""
        raise NotImplementedError

    def __get_item(self, system: str):
        """returns latest ContactPoint information of given system

        E.g. you can use Person.phone
        """
        # silently return nothing when given system is nonexistent.
        if system not in [key for key, value in ContactPoint.CONTACT_POINT_SYSTEM]:
            return ""

        items = self.telecom.filter(system=system)
        if items:
            # return first value that matches the given system
            return items.order_by("weight").first().value
        else:
            return ""

    @property
    def phone(self):
        return self.__get_item("phone")

    @property
    def email(self):
        return self.__get_item("email")


class Company(Organization):
    class Meta:
        verbose_name_plural = _("Companies")

    type = "other"


class InsuranceCompany(Company):
    """Either a private, or a social insurance company"""

    type = "ins"


class HealthServiceProvider(Organization):
    """A base class for Practitioner, Hospital/Department and other Health service providers."""

    type = "prov"


class Physician(HealthServiceProvider):
    """A Physician that is known to the user and can be referenced."""

    # name should be auto-generated from person

    type = "other"

    person = models.ForeignKey("Person", on_delete=models.CASCADE)
    speciality = models.ForeignKey(
        Specialty,
        on_delete=models.PROTECT,
        help_text=_("Medical field"),
        related_name="physicians",
    )
    personal_salutation = models.CharField(
        max_length=100,
        help_text=_(
            "if you know this physician personally, you can choose to customize his salutation, "
            "like 'Dear Thomas,'"
        ),
        blank=True,
    )

    def __str__(self):
        return f"{self.person}"


class Hospital(HealthServiceProvider):
    """A Hospital representation which has many departments.

    see :class:`Department`
    """

    def type(self):
        # not specified in FHIR
        return "prov"


class Department(PackageDataModel):
    """A department in a hospital"""

    hospital = models.ForeignKey(
        Hospital, on_delete=models.CASCADE, related_name="departments"
    )
    name = models.CharField(max_length=255)
    # shortcut = models.CharField(max_length=10)

    # Here you can only provide the direct extension of the department.
    # the phone number will be calculated automatically using the hospital's
    # using phone()
    telecom = models.ManyToManyField(ContactPoint)

    def phone(self):
        hospital_phone = self.hospital.telecom.objects.get(
            system="phone"
        )  # type: ContactPoint
        extension = self.telecom.objects.get(system="phone")
        if hospital_phone.value and extension.value:
            return f"{hospital_phone.value} - {extension.value}"

    def __str__(self):
        return f"{self.hospital.name} - {self.name}"

    def type(self):  # noQa
        return "dept"


class Name(CreatedModifiedModel):
    """All the names a person is known under"""

    # As opposed to the versioning of all other tables,
    # changed names should not be soft-deleted.
    # Search functionality must be available at any time for all names a
    # person ever had.

    last_name = models.CharField(max_length=255)
    first_name = models.CharField(max_length=255)
    preferred = models.CharField(
        max_length=255,
        help_text=_(
            "preferred first name, the name a person is usually called (nickname)"
        ),
        blank=True,
        null=True,
    )
    comment = models.CharField(
        max_length=255,
        help_text=_(
            "a comment regarding this name, useful in things like 'name before marriage' etc."
        ),
        blank=True,
        null=True,
    )
    weight = PositiveIntegerField(default=1)

    def __str__(self):
        # TODO: make order configurable
        return f"{self.last_name.upper()}, {self.first_name}"

    class Meta:
        ordering = ["weight"]


class Person(BaseModel):
    """A generic, natural person, or identity

    A person is mostly in relation with a patient, like his caregiver, etc."""

    # patient = models.ForeignKey(
    #     "Patient",
    #     blank=True,
    #     on_delete=models.PROTECT,
    #     help_text=_("The patient this person is related to."),
    # )
    # relationship = models.ForeignKey(
    #     RelationType,
    #     on_delete=models.PROTECT,
    #     help_text=_(
    #         "The relation this person has to the patient, e.g. Caregiver, Taxi drier etc."
    #     ),
    # )
    names = models.ManyToManyField(Name)
    title = models.CharField(
        max_length=50, blank=True, null=True, help_text=_("Academic title")
    )
    gender = models.ForeignKey(
        AdministrativeGender, blank=True, null=True, on_delete=models.PROTECT
    )
    birth_date = models.DateField(
        verbose_name=_("Date of birth"), null=True, blank=True
    )
    birth_time = models.TimeField(
        verbose_name=_("Time of birth"), blank=True, null=True
    )
    birth_date_is_estimated = models.BooleanField(default=False)

    # country of birth as per date of birth, coded as 2 character ISO code
    # TODO: make choices=ISO3166 fixture
    country_of_birth = models.ForeignKey(
        Country,
        help_text=_("ISO code of Country of Birth"),
        on_delete=models.SET_NULL,
        null=True,
    )

    # date when a person has died
    deceased_date = models.DateField(
        verbose_name=_("Date of death"), blank=True, null=True
    )
    deceased_time = models.TimeField(
        verbose_name=_("Time of death"), blank=True, null=True
    )
    deceased_is_estimated = models.BooleanField(default=False)

    addresses = models.ManyToManyField(Address, through="AddressMapper")

    # FHIR: telecom
    telecom = models.ManyToManyField(
        ContactPoint,
        help_text=_(
            "A contact detail for the person, e.g. a telephone number or an email address"
        ),
    )
    photo = models.ImageField(blank=True)
    active = models.BooleanField(
        help_text=_("Whether this person's record is in active use."), default=True
    )

    # # TODO: implement generatePubic and set required=True
    # pupic = models.CharField(
    #     max_length=24,
    #     blank=True,
    #     null=True,
    #     help_text=_(
    #         "Portable Unique Person Identification Code as per GNUmed white papers"
    #     ),
    # )
    #
    # def generate_pupic(self):
    #     """Generates a Portable Unique Person Identification according to
    #     GNUmed definition"""
    #     # TODO: implement PUPIC generation
    #     raise NotImplementedError

    @property
    def name(self):
        """Returns latest concatenated full name of Person's names list"""
        # return first in the list of names
        # TODO: add title here?
        return self.names.order_by("weight").first()

    @property
    def first_name(self):
        """Returns latest firstname in Person's names list"""
        return self.name.first_name

    @property
    def last_name(self):
        """Returns latest lastname in Person's names list"""
        return self.name.last_name

    def __get_item(self, system: str):
        """returns latest ContactPoint information of given system

        E.g. you can use Person.phone
        """
        # silently return nothing when given system is nonexistent.
        if system not in [key for key, value in ContactPoint.CONTACT_POINT_SYSTEM]:
            return ""

        items = self.telecom.filter(system=system)
        if items:
            # return first value that matches the given system
            return items.order_by("weight").first().value
        else:
            return ""

    @property
    def phone(self):
        return self.__get_item("phone")

    @property
    def email(self):
        return self.__get_item("email")

    def __str__(self):
        return f"{self.name}, {self.birth_date}"


class ContactPerson(Person):
    """A contact person within a company."""

    organization = models.ForeignKey(
        Organization,
        null=True,
        on_delete=models.SET_NULL,
        related_name="contact_persons",
    )

    def __str__(self):
        return f"{self.organization}: {self.name}"



from django.core.validators import RegexValidator

# fhir_server_allowed_references = "Account|ActivityDefinition|AdverseEvent|AllergyIntolerance|Appointment|" \
#         "AppointmentResponse|AuditEvent|Basic|Binary|BiologicallyDerivedProduct|" \
#         "BodyStructure|Bundle|CapabilityStatement|CarePlan|CareTeam|ChargeItem|Claim|" \
#         "ClaimResponse|ClinicalImpression|CodeSystem|Communication|CommunicationRequest|" \
#         "CompartmentDefinition|Composition|ConceptMap|Condition|Consent|Contract|" \
#         "Coverage|DetectedIssue|Device|DeviceComponent|DeviceMetric|DeviceRequest|" \
#         "DeviceUseStatement|DiagnosticReport|DocumentManifest|DocumentReference|" \
#         "EligibilityRequest|EligibilityResponse|Encounter|Endpoint|EnrollmentRequest|" \
#         "EnrollmentResponse|EntryDefinition|EpisodeOfCare|EventDefinition|" \
#         "ExampleScenario|ExpansionProfile|ExplanationOfBenefit|FamilyMemberHistory|Flag|" \
#         "Goal|GraphDefinition|Group|GuidanceResponse|HealthcareService|ImagingStudy|" \
#         "Immunization|ImmunizationEvaluation|ImmunizationRecommendation|" \
#         "ImplementationGuide|Invoice|ItemInstance|Library|Linkage|List|Location|" \
#         "Measure|MeasureReport|Media|Medication|MedicationAdministration|" \
#         "MedicationDispense|MedicationRequest|MedicationStatement|MedicinalProduct|" \
#         "MedicinalProductAuthorization|MedicinalProductClinicals|" \
#         "MedicinalProductDeviceSpec|MedicinalProductIngredient|" \
#         "MedicinalProductPackaged|MedicinalProductPharmaceutical|MessageDefinition|" \
#         "MessageHeader|NamingSystem|NutritionOrder|Observation|" \
#         "ObservationDefinition|OccupationalData|OperationDefinition|" \
#         "OperationOutcome|Organization|OrganizationRole|Patient|PaymentNotice|" \
#         "PaymentReconciliation|Person|PlanDefinition|Practitioner|" \
#         "PractitionerRole|Procedure|ProcessRequest|ProcessResponse|" \
#         "ProductPlan|Provenance|Questionnaire|QuestionnaireResponse|" \
#         "Relationship|RequestGroup|ResearchStudy|ResearchSubject|" \
#         "RiskAssessment|Schedule|SearchParameter|Sequence|" \
#         "ServiceRequest|Slot|Specimen|SpecimenDefinition|StructureDefinition|" \
#         "StructureMap|Subscription|Substance|SubstancePolymer|" \
#         "SubstanceReferenceInformation|SubstanceSpecification|SupplyDelivery|" \
#         "SupplyRequest|Task|TerminologyCapabilities|TestReport|TestScript|" \
#         "UserSession|ValueSet|VerificationResult|VisionPrescription" + "|Any"


CodeValidator = RegexValidator(
    regex=r"[^\s]+([\s]?[^\s]+)*", message="A Code must not contain whitespaces."
)

IdValidator = RegexValidator(
    regex=r"[a-z][A-Z][0-9][-.]?",
    message="Ids can be up to 64 characters long, "
    "and contain any combination of upper "
    "and lowercase ASCII letters, numerals,"
    ' "-" and "."',
)

OidValidator = RegexValidator(r"[0-2](\.[1-9]\d*)+", message="Given string is no OID")



import logging
import re
import mkdocs.plugins

log = logging.getLogger("mkdocs")


@mkdocs.plugins.event_priority(-50)
def on_page_markdown(markdown, page, **kwargs):
    """Finds non-https links"""
    path = page.file.src_uri
    for m in re.finditer(r"\bhttp://[^) ]+", markdown):
        log.warning(f"Documentation file '{path}' contains a non-HTTPS link: {m[0]}")



from enum import Enum

from django.apps import apps
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import gettext_lazy as _

from ..fields import CodeField
from medux.core.models import PackageDataModel


class Settings(PackageDataModel):
    """A simple scoped key-value store for application settings.

    This is not to be mixed up with Django settings.py. It should not be a replacement for that, there are plenty oft
    projects that deal with that sort of settings. It's meant as an application configuration method that could be
    overridden by a vendor, settings that could be set per workspace, preferences per user, etc.

    Each setting has a scope which is important to be mentioned. There are different settings scopes:

    global
      Settings that are defaults to MedUX. Each app can add settings to that scope.

    vendor
      These are settings that are provided by the vendor. Only vendor settings should be packaged in :ref:`Datapack`s.

    system
      These settings are specific to one "appliance" or "installation" of MedUX.

    machine
      applicable for one computer, or workplace.

    user
      Settings that only affect one user.

    These settings are searched for and applied in that order. So, if you retrieve a setting for
    Settings can be packaged as :ref:`Datapack`, and hence updated by the vendor (Here, only "vendor" scoped settings
    should be provided via datapacks!)

    The ``key`` must be a dotted path, starting with the name of an app, followed by a dot, like `medux.foo_setting`.
    """

    class Scope(Enum):
        GLOBAL = "global"
        VENDOR = "vendor"
        SYSTEM = "system"
        MACHINE = "machine"
        USER = "user"

    scope = CodeField(terminology_binding=Scope)
    key = models.CharField(max_length=255)
    value = models.CharField(max_length=255, blank=True)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text=_(
            "The user this setting is scoped to. Just applicable if scope == 'user'."
        ),
    )

    class Meta:
        unique_together = ("scope", "key")

    def __str__(self):
        return f"{self.key} ({self.scope})"

    def clean(self):
        if self.scope == Settings.Scope.USER and self.user is None:
            raise ValidationError(
                {"title": _("'User' must be given when scope == 'USER'.")}
            )
        if not self.key.startswith(
            tuple([key + "." for key in apps.app_configs.keys()])
        ):
            raise ValidationError(
                _("'Key' must start with an app name (followed by a dot)")
            )
        for part in self.key.split("."):
            if not part.isidentifier():
                raise ValidationError(
                    _(
                        "Key must be dotted valid python identifiers, like 'medux.foo_setting.baz'."
                    )
                )

    def save(self, *args, **kwargs):
        self.full_clean()
        self.scope = self.scope.value
        super().save(*args, **kwargs)

    @staticmethod
    def get(key: str, default="", scope=None) -> str:
        """Retrieves the current valid value for ``key``, matching the scope.

        :param key: the dotted settings parameter to retrieve a value for. Must start with
                    a name of an app, followed by a dot, like 'medux.foo_setting'.
        :param scope: if given, only the setting with the given scope is returned.
                      If not, the highest ranked scope is used.
        :param default: if setting is not found, this default will be returned.
        """

        settings = Settings.objects.filter(key=key).order_by("scope")
        if scope:
            settings = settings.filter(scope=scope)
            # FIXME: make sure here that order_by orders correctly - NOT alphabetically!

        settings = settings.last()
        if settings is None:
            return str(default)
        return settings.value



from django import forms
from django.utils.translation import gettext_lazy as _

from medux.core.models import Dosage, MedicationRequest


class QuantityField(forms.CharField):
    widget = forms.widgets.TextInput


class DosageWidget(forms.TextInput):
    pass


class DosageForm(forms.ModelForm):
    class Meta:
        model = Dosage
        widgets = {"dose_rate_numerator_value": DosageWidget}
        fields = [
            "text",
            "additional_instructions",
            "patient_instructions",
            "as_needed",
            "route",
            "method",
            "dose_type",
            "dose_quantity_value",
            "dose_quantity_unit",
            # "dose_rate_nominator_value",
        ]


class CreatePrescriptionForm:
    model = MedicationRequest
